from internetarchive.cli.cli_utils import get_args_dict


def test_get_args_dict():
    test_input = [
        'collection:test_collection',
        'description: Attention: multiple colons',
        'unicode_test:தமிழ்',
        'subject:subject1, subject1',
        'subject:subject2',
        'subject:subject3; subject3',
    ]
    test_output = {
        'collection': 'test_collection',
        'description': ' Attention: multiple colons',
        'unicode_test': 'தமிழ்',
        'subject': ['subject1, subject1', 'subject2', 'subject3; subject3'],
    }
    args_dict = get_args_dict(test_input)
    for key, value in args_dict.items():
        assert test_output[key] == value


def test_get_args_dict_query_string():
    test_input = ['a=b,foo&c=d&e=f', 'foo:bar ']
    test_output = {
        'a': 'b,foo',
        'c': 'd',
        'e': 'f',
        'foo': 'bar ',
    }
    args_dict = get_args_dict(test_input, query_string=True)
    for key, value in args_dict.items():
        assert test_output[key] == value



import responses

from internetarchive.utils import json
from tests.conftest import PROTOCOL, IaRequestsMock, ia_call, load_test_data_file


def test_ia_search_itemlist(capsys):
    test_scrape_response = load_test_data_file('scrape_response.json')

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        url = f'{PROTOCOL}//archive.org/services/search/v1/scrape'
        p1 = {
            'q': 'collection:attentionkmartshoppers',
            'count': '10000'
        }
        _j = json.loads(test_scrape_response)
        del _j['cursor']
        _r = json.dumps(_j)
        rsps.add(responses.POST, url,
                 body=_r,
                 match=[responses.matchers.query_param_matcher(p1)])
        ia_call(['ia', 'search', 'collection:attentionkmartshoppers', '--itemlist'])

    out, err = capsys.readouterr()
    assert len(set(out.split())) == 100


def test_ia_search_num_found(capsys):
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        url = f'{PROTOCOL}//archive.org/services/search/v1/scrape'
        p = {
            'q': 'collection:nasa',
            'total_only': 'true',
            'count': '10000'
        }
        rsps.add(responses.POST, url,
                 body='{"items":[],"count":0,"total":50}',
                 match=[responses.matchers.query_param_matcher(p)])

        ia_call(['ia', 'search', 'collection:nasa', '--num-found'])
    out, err = capsys.readouterr()
    assert out == '50\n'



"""
interneratchive.cli.cli_utils

"""

# Copyright (C) 2012-2024 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import argparse
import json
import os
import signal
import sys
from collections import defaultdict
from collections.abc import Iterable
from typing import Mapping
from urllib.parse import parse_qsl

from internetarchive.utils import InvalidIdentifierException, validate_s3_identifier


def get_args_dict(args: list[str],
                  query_string: bool = False,
                  header: bool = False) -> dict:
    args = args or []
    if not isinstance(args, list):
        args = [args]
    metadata: dict[str, list | str] = defaultdict(list)
    for md in args:
        if query_string:
            if (":" in md) and ("=" not in md):
                md = md.replace(":", "=").replace(";", "&")
            for key, value in parse_qsl(md):
                assert value
                metadata[key] = value
        else:
            key, value = md.split(":", 1)
            assert value
            if value not in metadata[key]:
                metadata[key].append(value)  # type: ignore

    for key in metadata:  # noqa: PLC0206
        # Flatten single item lists.
        if len(metadata[key]) <= 1:
            metadata[key] = metadata[key][0]

    return metadata


def convert_str_list_to_unicode(str_list: list[bytes]):
    encoding = sys.getfilesystemencoding()
    return [b.decode(encoding) for b in str_list]


def validate_identifier(identifier):
    try:
        validate_s3_identifier(identifier)
    except InvalidIdentifierException as e:
        raise argparse.ArgumentTypeError(str(e))
    return identifier


def flatten_list(lst):
    """Flatten a list if it contains lists."""
    result = []
    for item in lst:
        if isinstance(item, Iterable) and not isinstance(item, str):
            result.extend(flatten_list(item))  # Recursively flatten
        else:
            result.append(item)  # Just append the item if it's not a list
    return result


class FlattenListAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        # Flatten the list of values (if nested)
        flattened = flatten_list(values)
        # Initialize the attribute if it doesn't exist yet
        if getattr(namespace, self.dest, None) is None:
            setattr(namespace, self.dest, [])
        # Append the flattened list to the existing attribute
        getattr(namespace, self.dest).extend(flattened)


class PostDataAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        current_value = getattr(namespace, self.dest, None)

        # Split values into individual JSON objects (if needed) and parse them
        all_values = []
        for value in values:
            try:
                obj = json.loads(value)
                all_values.append(obj)
            except json.JSONDecodeError as e:
                parser.error(f"Invalid JSON format for post data: {value}")

        # If there is no current value (first argument), initialize it as an object or list
        if current_value is None:
            # If there's only one value, don't wrap it in a list
            if len(all_values) == 1:
                post_data = all_values[0]
            else:
                post_data = all_values
        elif isinstance(current_value, list):
            # If it's already a list, append the new values to it
            post_data = current_value + all_values
        else:
            # If it's a single object (first argument), convert it into a list and append new data
            post_data = [current_value] + all_values

        # Set the final value back to the namespace
        setattr(namespace, self.dest, post_data)


class QueryStringAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        # Initialize the destination as an empty dictionary if it doesn't exist
        if getattr(namespace, self.dest, None) is None:
            setattr(namespace, self.dest, {})

        for sublist in values:
            if "=" not in sublist and ":" in sublist:
                sublist = sublist.replace(":", "=", 1)
            key_value_pairs = parse_qsl(sublist)

            if sublist and not key_value_pairs:
                parser.error(f"{option_string} must be formatted as 'key=value' "
                              "or 'key:value'")

            for key, value in key_value_pairs:
                current_dict = getattr(namespace, self.dest)
                if key in current_dict:
                    current_dict[key].append(value)
                else:
                    current_dict[key] = [value]

        current_dict = getattr(namespace, self.dest)
        for key, value in current_dict.items():
            if len(value) == 1:
                current_dict[key] = value[0]


class MetadataAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        # Initialize the destination as an empty dictionary if it doesn't exist
        if getattr(namespace, self.dest, None) is None:
            setattr(namespace, self.dest, {})

        for sublist in values:
            if ":" not in sublist and "=" in sublist:
                sublist = sublist.replace("=", ":", 1)
            try:
                key, value = sublist.split(":", 1)
            except ValueError:
                parser.error(f"{option_string} must be formatted as 'KEY:VALUE'")

            current_dict = getattr(namespace, self.dest)
            if key in current_dict:
                if not isinstance(current_dict[key], list):
                    current_dict[key] = [current_dict[key]]
                current_dict[key].append(value)
            else:
                current_dict[key] = value


def validate_dir_path(path):
    """
    Check if the given path is a directory that exists.

    Args:
        path (str): The path to check.

    Returns:
        str: The validated directory path.

    Raises:
        argparse.ArgumentTypeError: If the path is not a valid directory.
    """
    if os.path.isdir(path):
        return path
    else:
        raise argparse.ArgumentTypeError(f"'{path}' is not a valid directory")


def exit_on_signal(sig, frame):
    """
    Exit the program cleanly upon receiving a specified signal.

    This function is designed to be used as a signal handler. When a signal
    (such as SIGINT or SIGPIPE) is received, it exits the program with an
    exit code of 128 plus the signal number. This convention helps to
    distinguish between regular exit codes and those caused by signals.
    """
    exit_code = 128 + sig
    sys.exit(exit_code)



#
# The internetarchive module is a Python/CLI interface to Archive.org.
#
# Copyright (C) 2012-2024 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
internetarchive.iarequest
~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (C) 2012-2025 by Internet Archive.
:license: AGPL 3, see LICENSE for more details.
"""
import copy
import logging
import re
from urllib.parse import quote

import requests
import requests.models
from jsonpatch import make_patch

from internetarchive import __version__, auth
from internetarchive.exceptions import ItemLocateError
from internetarchive.utils import delete_items_from_dict, json, needs_quote

logger = logging.getLogger(__name__)


class S3Request(requests.models.Request):
    def __init__(self,
                 metadata=None,
                 file_metadata=None,
                 queue_derive=True,
                 access_key=None,
                 secret_key=None,
                 set_scanner=True,
                 **kwargs):

        super().__init__(**kwargs)

        self.auth = self.auth or auth.S3Auth(access_key, secret_key)
        self.metadata = metadata or {}
        self.file_metadata = file_metadata or {}
        self.queue_derive = queue_derive
        self.set_scanner = set_scanner

    def prepare(self):
        p = S3PreparedRequest()
        p.prepare(
            method=self.method,
            url=self.url,
            headers=self.headers,
            files=self.files,
            data=self.data,
            params=self.params,
            auth=self.auth,
            cookies=self.cookies,
            hooks=self.hooks,

            # S3Request kwargs.
            metadata=self.metadata,
            file_metadata=self.file_metadata,
            queue_derive=self.queue_derive,
            set_scanner=self.set_scanner,
        )
        return p


class S3PreparedRequest(requests.models.PreparedRequest):
    def prepare(self, method=None, url=None, headers=None, files=None, data=None,
                params=None, auth=None, cookies=None, hooks=None, queue_derive=None,
                metadata=None, file_metadata=None, set_scanner=None):
        self.prepare_method(method)
        self.prepare_url(url, params)
        self.prepare_headers(headers, metadata, file_metadata, queue_derive, set_scanner)
        self.prepare_cookies(cookies)
        self.prepare_body(data, files)
        self.prepare_auth(auth, url)
        # Note that prepare_auth must be last to enable authentication schemes
        # such as OAuth to work on a fully prepared request.

        # This MUST go after prepare_auth. Authenticators could add a hook
        self.prepare_hooks(hooks)

    def prepare_headers(self, headers, metadata, file_metadata, queue_derive,
                        set_scanner):
        headers = headers.copy() if headers else {}
        metadata = metadata.copy() if metadata else {}
        file_metadata = file_metadata.copy() if file_metadata else {}

        if set_scanner:
            scanner_value = f'Internet Archive Python library {__version__}'
            existing_scanner = metadata.get('scanner', [])
            if not isinstance(existing_scanner, list):
                existing_scanner = [existing_scanner]
            existing_scanner.append(scanner_value)
            metadata['scanner'] = existing_scanner
        prepared_metadata = prepare_metadata(metadata)
        prepared_file_metadata = prepare_metadata(file_metadata)

        headers.setdefault('x-archive-auto-make-bucket', '1')
        headers['x-archive-queue-derive'] = '0' if queue_derive is False else '1'

        self._add_metadata_headers(headers, prepared_metadata, 'meta')
        self._add_metadata_headers(headers, prepared_file_metadata, 'filemeta')

        super().prepare_headers(headers)

    def _add_metadata_headers(self, headers, prepared_metadata, meta_type):
        for key, values in prepared_metadata.items():
            if not isinstance(values, list):
                values = [values]
            for idx, value in enumerate(values):
                if not value:
                    continue
                header_key = f'x-archive-{meta_type}{idx:02d}-{key}'.replace('_', '--')
                if isinstance(value, str) and needs_quote(value):
                    value = f'uri({quote(value)})'
                headers[header_key] = value


class MetadataRequest(requests.models.Request):
    def __init__(self,
                 metadata=None,
                 source_metadata=None,
                 target=None,
                 priority=None,
                 access_key=None,
                 secret_key=None,
                 append=None,
                 expect=None,
                 append_list=None,
                 insert=None,
                 reduced_priority=None,
                 **kwargs):

        super().__init__(**kwargs)

        self.auth = self.auth or auth.S3PostAuth(access_key, secret_key)
        self.metadata = metadata or {}
        self.source_metadata = source_metadata
        self.target = target
        self.priority = priority
        self.append = append
        self.expect = expect or {}
        self.append_list = append_list
        self.insert = insert
        self.reduced_priority = reduced_priority

    def prepare(self):
        p = MetadataPreparedRequest()
        p.prepare(
            method=self.method,
            url=self.url,
            headers=self.headers,
            files=self.files,
            data=self.data,
            params=self.params,
            auth=self.auth,
            cookies=self.cookies,
            hooks=self.hooks,

            # MetadataRequest kwargs.
            metadata=self.metadata,
            priority=self.priority,
            source_metadata=self.source_metadata,
            target=self.target,
            append=self.append,
            expect=self.expect,
            append_list=self.append_list,
            insert=self.insert,
            reduced_priority=self.reduced_priority,
        )
        return p


class MetadataPreparedRequest(requests.models.PreparedRequest):
    def prepare(self, method=None, url=None, headers=None, files=None, data=None,
                params=None, auth=None, cookies=None, hooks=None, metadata=None,
                source_metadata=None, target=None, priority=None, append=None,
                expect=None, append_list=None, insert=None, reduced_priority=None):
        # First handle our custom headers
        if reduced_priority:
            headers = headers.copy() if headers else {}
            headers['X-Accept-Reduced-Priority'] = '1'

        # Now run full parent preparation
        super().prepare(
            method=method,
            url=url,
            headers=headers,
            files=files,
            data=data,
            params=params,
            auth=auth,
            cookies=cookies,
            hooks=hooks,
        )

        # Now add our custom handling
        self.identifier = self.url.split('?')[0].split('/')[-1]
        self._prepare_request_body(
            metadata,
            source_metadata,
            target,
            priority,
            append,
            append_list,
            insert,
            expect,
        )
        self.prepare_auth(auth, url)
        # Note that prepare_auth must be last to enable authentication schemes
        # such as OAuth to work on a fully prepared request.

        # This MUST go after prepare_auth. Authenticators could add a hook
        self.prepare_hooks(hooks)

    def _prepare_request_body(self, metadata, source_metadata, target, priority,
                              append, append_list, insert, expect):
        if not source_metadata:
            r = requests.get(self.url, timeout=10)
            source_metadata = r.json()

        if self._is_multi_target(metadata):
            changes = self._prepare_multi_target_changes(
                metadata,
                source_metadata,
                target,
                append,
                expect,
                append_list,
                insert,
            )
            self.data = {'-changes': json.dumps(changes), 'priority': priority or -5}
        else:
            self._prepare_single_target_body(
                metadata,
                source_metadata,
                target,
                append,
                append_list,
                insert,
                expect,
                priority,
            )

        logger.debug(f'submitting metadata request: {self.data}')
        super().prepare_body(self.data, None)

    def _is_multi_target(self, metadata):
        return (
            isinstance(metadata, list)
            or any('/' in k for k in metadata)
            or all(isinstance(v, dict) for v in metadata.values())
        )

    def _prepare_multi_target_changes(self, metadata, source_metadata, target,
                                      append, expect, append_list, insert):
        changes = []
        if target:
            metadata = {target: metadata}
        for key in metadata:
            patch = self._get_patch_for_target(
                key,
                metadata[key],
                source_metadata,
                append,
                expect,
                append_list,
                insert,
            )
            changes.append({'target': key, 'patch': patch})
        return changes

    def _prepare_single_target_body(self, metadata, source_metadata, target, append,
                                    append_list, insert, expect, priority):
        target = target or 'metadata'
        if target == 'metadata':
            try:
                patch = prepare_patch(
                    metadata,
                    source_metadata['metadata'],
                    append,
                    expect,
                    append_list,
                    insert,
                )
            except KeyError:
                raise ItemLocateError(
                    f'{self.identifier} cannot be located '
                    'because it is dark or does not exist.'
                )
        elif target.startswith('files/'):
            patch = prepare_files_patch(
                metadata,
                source_metadata['files'],
                target,
                append,
                append_list,
                insert,
                expect,
            )
        else:
            patch = prepare_target_patch(
                {target: metadata},
                source_metadata,
                append,
                target,
                append_list,
                target,
                insert,
                expect,
            )
        self.data = {
            '-patch': json.dumps(patch),
            '-target': target,
            'priority': priority or -5,
        }


def prepare_patch(metadata, source_metadata, append, expect=None,
                  append_list=None, insert=None):
    destination = source_metadata.copy()
    if isinstance(metadata, list):
        prepared_metadata = metadata
        if not destination:
            destination = []
    else:
        prepared_metadata = prepare_metadata(
            metadata,
            source_metadata,
            append,
            append_list,
            insert,
        )
    if isinstance(destination, dict):
        destination.update(prepared_metadata)
    elif isinstance(metadata, list):
        destination = prepared_metadata if not destination else prepared_metadata
    else:
        if isinstance(prepared_metadata, list):
            destination = prepared_metadata
        else:
            destination = [prepared_metadata]

    destination = delete_items_from_dict(destination, 'REMOVE_TAG')
    patch = make_patch(source_metadata, destination).patch
    patch_tests = _create_patch_tests(expect)
    return patch_tests + patch


def _create_patch_tests(expect):
    tests = []
    for key, value in (expect or {}).items():
        if '[' in key:
            parts = key.split('[')
            idx = int(parts[1].strip(']'))
            path = f'/{parts[0]}/{idx}'
        else:
            path = f'/{key}'
        tests.append({'op': 'test', 'path': path, 'value': value})
    return tests


def prepare_target_patch(metadata, source_metadata, append, target,
                         append_list, key, insert, expect):
    nested_dict = _create_nested_dict(metadata)
    current = source_metadata
    for part in key.split('/'):
        current = current.get(part, {})
    patch = prepare_patch(nested_dict, current, append, expect, append_list, insert)
    return patch


def _create_nested_dict(metadata):
    nested = {}
    for key_path, value in metadata.items():
        parts = key_path.split('/')
        current = nested
        for part in parts[:-1]:
            current = current.setdefault(part, {})
        current[parts[-1]] = value
    return nested


def prepare_files_patch(metadata, files_metadata, target, append,
                        append_list, insert, expect):
    filename = target.split('/')[1]
    for file_meta in files_metadata:
        if file_meta.get('name') == filename:
            return prepare_patch(
                metadata,
                file_meta,
                append,
                expect,
                append_list,
                insert,
            )
    return []


def prepare_metadata(metadata, source_metadata=None, append=False,
                     append_list=False, insert=False):
    source = copy.deepcopy(source_metadata) if source_metadata else {}
    prepared = {}

    indexed_keys = _process_indexed_keys(metadata, source, prepared)
    _process_non_indexed_keys(metadata, source, prepared, append, append_list, insert)
    _cleanup_indexed_keys(prepared, indexed_keys, metadata)

    return prepared


def _process_non_indexed_keys(metadata, source, prepared, append, append_list, insert):
    for key, value in metadata.items():
        current_key = key

        if isinstance(value, (int, float, complex)) and not isinstance(value, bool):
            value = str(value)

        if append_list and source.get(current_key):
            existing = source[current_key]
            if not isinstance(existing, list):
                existing = [existing]
            prepared[current_key] = existing + [value]
        elif append and source.get(current_key):
            prepared[current_key] = f'{source[current_key]} {value}'
        elif insert and source.get(current_key):
            existing = source[current_key]
            if not isinstance(existing, list):
                existing = [existing]
            existing.insert(0, value)
            prepared[current_key] = [v for v in existing if v]
        else:
            prepared[current_key] = value


def _cleanup_indexed_keys(prepared, indexed_keys, metadata):
    for base in indexed_keys:
        if base in prepared:
            prepared[base] = [v for v in prepared[base] if v is not None]
            indexes = [
                i for i, k in enumerate(metadata)
                if _get_base_key(k) == base and metadata[k] == 'REMOVE_TAG'
            ]
            for i in reversed(indexes):
                if i < len(prepared[base]):
                    del prepared[base][i]


def _process_indexed_keys(metadata, source, prepared):
    indexed_keys = {}
    for key in list(metadata.keys()):
        if _is_indexed_key(key):
            base = _get_base_key(key)
            idx = _get_index(key)

            if base not in indexed_keys:
                source_list = source.get(base, [])
                if not isinstance(source_list, list):
                    source_list = [source_list]
                indexed_keys[base] = len(source_list)

                current_metadata_length = len(metadata)
                prepared[base] = source_list + [None] * (
                    current_metadata_length - len(source_list)
                )

            while len(prepared[base]) <= idx:
                prepared[base].append(None)

            prepared[base][idx] = metadata[key]
            del metadata[key]
    return indexed_keys


def _get_base_key(key):
    return key.split('[')[0]


def _is_indexed_key(key):
    return '[' in key and ']' in key


def _get_index(key):
    match = re.search(r'(?<=\[)\d+(?=\])', key)
    return int(match.group()) if match else None



"""
ia_tasks.py

'ia' subcommand for retrieving information about archive.org catalog tasks.
"""

# Copyright (C) 2012-2024 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import sys
import warnings

from internetarchive.cli.cli_utils import PostDataAction, QueryStringAction
from internetarchive.utils import json


def setup(subparsers):
    """
    Setup args for tasks command.

    Args:
        subparsers: subparser object passed from ia.py
    """
    parser = subparsers.add_parser("tasks",
                                   aliases=["ta"],
                                   help="Retrieve information about your archive.org catalog tasks")

    parser.add_argument("-t", "--task",
                        nargs="*",
                        help="Return information about the given task.")
    parser.add_argument("-G", "--get-task-log",
                        help="Return the given tasks task log.")
    parser.add_argument("-p", "--parameter",
                        nargs="+",
                        action=QueryStringAction,
                        default={},
                        metavar="KEY:VALUE",
                        help="URL parameters passed to catalog.php.")
    parser.add_argument("-T", "--tab-output",
                        action="store_true",
                        help="Output task info in tab-delimited columns.")
    parser.add_argument("-c", "--cmd",
                        type=str,
                        help="The task to submit (e.g., make_dark.php).")
    parser.add_argument("-C", "--comment",
                        type=str,
                        help="A reasonable explanation for why a task is being submitted.")
    parser.add_argument("-a", "--task-args",
                        nargs="+",
                        action=QueryStringAction,
                        default={},
                        metavar="KEY:VALUE",
                        help="Args to submit to the Tasks API.")
    parser.add_argument("-d", "--data",
                        nargs="+",
                        action=PostDataAction,
                        metavar="KEY:VALUE",
                        default={},
                        help="Additional data to send when submitting a task.")
    parser.add_argument("-r", "--reduced-priority",
                        action="store_true",
                        help="Submit task at a reduced priority.")
    parser.add_argument("-l", "--get-rate-limit",
                        action="store_true",
                        help="Get rate limit info.")
    parser.add_argument("identifier",
                        type=str,
                        nargs="?",
                        help="Identifier for tasks specific operations.")

    parser.set_defaults(func=lambda args: main(args, parser))


def handle_task_submission_result(result, cmd):
    """
    Handle the result of a task submission.
    """
    if result.get("success"):
        task_log_url = result.get("value", {}).get("log")
        print(f"success: {task_log_url}", file=sys.stderr)
    elif "already queued/running" in result.get("error", ""):
        print(f"success: {cmd} task already queued/running", file=sys.stderr)
    else:
        print(f"error: {result.get('error')}", file=sys.stderr)
    sys.exit(0 if result.get("success") else 1)


def main(args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
    """
    Main entry point for 'ia tasks'.
    """
    # Tasks write API.
    if args.cmd:
        if args.get_rate_limit:
            r = args.session.get_tasks_api_rate_limit(args.cmd)
            print(json.dumps(r))
            sys.exit(0)
        args.data["args"] = args.task_args
        r = args.session.submit_task(args.identifier,
                                     args.cmd,
                                     comment=args.comment,
                                     priority=int(args.data.get("priority", 0)),
                                     reduced_priority=args.reduced_priority,
                                     data=args.data)
        handle_task_submission_result(r.json(), args.cmd)
        sys.exit(0)

    # Tasks read API.
    if args.identifier:
        _params = {"identifier": args.identifier, "catalog": 1, "history": 1}
        _params.update(args.parameter)
        args.parameter = _params
    elif args.get_task_log:
        log = args.session.get_task_log(args.get_task_log, **args.parameter)
        print(log.encode("utf-8", errors="surrogateescape")
                 .decode("utf-8", errors="replace"))
        sys.exit(0)

    queryable_params = [
        "identifier",
        "task_id",
        "server",
        "cmd",
        "args",
        "submitter",
        "priority",
        "wait_admin",
        "submittime",
    ]

    if not (args.identifier
            or args.parameter.get("task_id")):
        _params = {"catalog": 1, "history": 0}
        _params.update(args.parameter)
        args.parameter = _params

    if not any(x in args.parameter for x in queryable_params):
        _params = {"submitter": args.session.user_email, "catalog": 1, "history": 0, "summary": 0}
        _params.update(args.parameter)
        args.parameter = _params

    if args.tab_output:
        warn_msg = ("tab-delimited output will be removed in a future release. "
                    "Please switch to the default JSON output.")
        warnings.warn(warn_msg, stacklevel=2)
    for t in args.session.get_tasks(params=args.parameter):
        # Legacy support for tab-delimited output.
        # Mypy is confused by CatalogTask members being created from kwargs
        if args.tab_output:
            color = t.color if t.color else "done"
            task_args = "\t".join([f"{k}={v}" for k, v in t.args.items()])  # type: ignore
            output = "\t".join([str(x) for x in [
                t.identifier,
                t.task_id,
                t.server,
                t.submittime,
                t.cmd,
                color,
                t.submitter,
                task_args,
            ] if x])
            print(output, flush=True)
        else:
            print(t.json(), flush=True)



"""
ia_delete.py

'ia' subcommand for deleting files from archive.org items.
"""

# Copyright (C) 2012-2024 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import sys

import requests.exceptions

from internetarchive.cli.cli_utils import (
    FlattenListAction,
    MetadataAction,
    QueryStringAction,
    validate_identifier,
)
from internetarchive.utils import get_s3_xml_text


def setup(subparsers):
    """
    Setup args for delete command.

    Args:
        subparsers: subparser object passed from ia.py
    """
    parser = subparsers.add_parser("delete",
                                   aliases=["rm"],
                                   help="Delete files from archive.org items")
    # Positional arguments
    parser.add_argument("identifier",
                        type=validate_identifier,
                        help="Identifier for the item from which files are to be deleted.")
    parser.add_argument("file",
                        type=str,
                        nargs="*",
                        help="Specific file(s) to delete.")

    # Optional arguments
    parser.add_argument("-q", "--quiet",
                        action="store_true",
                        help="Print status to stdout.")
    parser.add_argument("-c", "--cascade",
                        action="store_true",
                        help="Delete all associated files including derivatives and the original.")
    parser.add_argument("-H", "--header",
                        nargs="+",
                        action=QueryStringAction,
                        default={},
                        metavar="KEY:VALUE",
                        help="S3 HTTP headers to send with your request.")
    parser.add_argument("-a", "--all",
                        action="store_true",
                        help="Delete all files in the given item. Some files cannot be deleted.")
    parser.add_argument("-d", "--dry-run",
                        action="store_true",
                        help=("Output files to be deleted to stdout, "
                              "but don't actually delete them."))
    parser.add_argument("-g", "--glob",
                        type=str,
                        help="Only delete files matching the given pattern.")
    parser.add_argument("-f", "--format",
                        type=str,
                        nargs="+",
                        action=FlattenListAction,
                        help="Only delete files matching the specified formats.")
    parser.add_argument("-R", "--retries",
                        type=int,
                        default=2,
                        help="Number of retries on S3 503 SlowDown error.")
    parser.add_argument("--no-backup",
                        action="store_true",
                        help="Turn off archive.org backups. Clobbered files will not be saved.")

    parser.set_defaults(func=lambda args: main(args, parser))


def get_files_to_delete(args: argparse.Namespace, item) -> list:
    """Get files to delete based on command-line arguments."""
    if args.all:
        files = list(item.get_files())
        args.cascade = True
    elif args.glob:
        files = item.get_files(glob_pattern=args.glob)
    elif args.format:
        files = item.get_files(formats=args.format)
    else:
        fnames = [f.strip() for f in (sys.stdin if args.file == ["-"] else args.file)]
        files = list(item.get_files(fnames))
    return files


def delete_files(files, args, item, verbose):
    """
    Deletes files from an item.

    Args:
        files (list): A list of files to delete.
        args (argparse.Namespace): Parsed command-line arguments.
        item: The item from which files are being deleted.
        verbose (bool): If True, verbose output is enabled.

    Returns:
        bool: True if errors occurred during deletion, False otherwise.
    """
    errors = False

    # Files that cannot be deleted via S3.
    no_delete = ["_meta.xml", "_files.xml", "_meta.sqlite"]

    for f in files:
        if not f:
            if verbose:
                print(f" error: '{f.name}' does not exist", file=sys.stderr)
            errors = True
            continue
        if any(f.name.endswith(s) for s in no_delete):
            continue
        if args.dry_run:
            print(f" will delete: {item.identifier}/{f.name}", file=sys.stderr)
            continue
        try:
            resp = f.delete(verbose=verbose,
                            cascade_delete=args.cascade,
                            headers=args.header,
                            retries=args.retries)
        except requests.exceptions.RetryError:
            print(f" error: max retries exceeded for {f.name}", file=sys.stderr)
            errors = True
            continue

        if resp.status_code != 204:
            errors = True
            msg = get_s3_xml_text(resp.content)
            print(f" error: {msg} ({resp.status_code})", file=sys.stderr)
            continue
    return errors


def main(args: argparse.Namespace, parser: argparse.ArgumentParser):
    """
    Main entry point for 'ia delete'.
    """
    verbose = not args.quiet
    item = args.session.get_item(args.identifier)
    if not item.exists:
        print(f"{item.identifier}: skipping, item doesn't exist.", file=sys.stderr)
        return

    # Add keep-old-version by default.
    if "x-archive-keep-old-version" not in args.header and not args.no_backup:
        args.header["x-archive-keep-old-version"] = "1"

    if verbose:
        print(f"Deleting files from {item.identifier}", file=sys.stderr)

    files = get_files_to_delete(args, item)

    if not files:
        print(" warning: no files found, nothing deleted.", file=sys.stderr)
        sys.exit(1)

    errors = delete_files(files, args, item, verbose)

    if errors:
        sys.exit(1)



import os
import re
import sys
from subprocess import PIPE, Popen

import pytest
import responses
from responses import RequestsMock

from internetarchive import get_session
from internetarchive.api import get_item
from internetarchive.cli import ia
from internetarchive.utils import json

PROTOCOL = 'https:'
BASE_URL = 'https://archive.org/'
METADATA_URL = f'{BASE_URL}metadata/'
ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
TEST_CONFIG = os.path.join(ROOT_DIR, 'tests/ia.ini')
NASA_METADATA_PATH = os.path.join(ROOT_DIR, 'tests/data/metadata/nasa.json')
NASA_EXPECTED_FILES = {
    'globe_west_540.jpg',
    'globe_west_540_thumb.jpg',
    'nasa_archive.torrent',
    'nasa_meta.sqlite',
    'nasa_files.xml',
    'nasa_meta.xml',
    'nasa_reviews.xml',
    'nasa_itemimage.jpg',
    '__ia_thumb.jpg',
}


def ia_call(argv, expected_exit_code=0):
    # Use a test config for all `ia` tests.
    argv.insert(1, '--config-file')
    argv.insert(2, TEST_CONFIG)
    sys.argv = argv
    try:
        ia.main()
    except SystemExit as exc:
        exit_code = exc.code if exc.code else 0
        assert exit_code == expected_exit_code


def files_downloaded(path):
    found_files = set()
    try:
        found_files = set(os.listdir(path))
    except OSError:
        pass
    return found_files


def load_file(filename):
    with open(filename) as fh:
        return fh.read()


def load_test_data_file(filename):
    return load_file(os.path.join(ROOT_DIR, 'tests/data/', filename))


def call_cmd(cmd, expected_exit_code=0):
    proc = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE)  # noqa: S602
    stdout, stderr = proc.communicate()
    stdout = stdout.decode('utf-8').strip()
    stderr = stderr.decode('utf-8').strip()
    if proc.returncode != expected_exit_code:
        print(stdout)
        print(stderr)
        assert proc.returncode == expected_exit_code
    return (stdout, stderr)


class IaRequestsMock(RequestsMock):
    def add_metadata_mock(self, identifier, body=None, method=responses.GET,
                          protocol='https?', transform_body=None):
        url = re.compile(f'{protocol}://archive.org/metadata/{identifier}')
        if body is None:
            body = load_test_data_file(f'metadata/{identifier}.json')
        if transform_body:
            body = transform_body(body)
        self.add(method, url, body=body, content_type='application/json')

    def mock_all_downloads(self, num_calls=1, body='test content', protocol='https?'):
        url = re.compile(f'{protocol}://archive.org/download/.*')
        for _ in range(6):
            self.add(responses.GET, url, body=body)


@pytest.fixture
def tmpdir_ch(tmpdir):
    tmpdir.chdir()
    return tmpdir


@pytest.fixture
def nasa_mocker():
    with IaRequestsMock() as mocker:
        mocker.add_metadata_mock('nasa')
        yield mocker


@pytest.fixture
def nasa_item():
    session = get_session()
    with IaRequestsMock() as mocker:
        mocker.add_metadata_mock('nasa')
        yield session.get_item('nasa')


@pytest.fixture
def session():
    return get_session(config={'s3': {'access': 'access', 'secret': 'secret'}})


@pytest.fixture
def nasa_metadata():
    return json.loads(load_test_data_file('metadata/nasa.json'))


# TODO: Why is this function defined twice in this file?  See issue #505
@pytest.fixture  # type: ignore
def nasa_item(nasa_mocker):  # noqa: F811
    return get_item('nasa')



"""
ia_metadata.py

'ia' subcommand for modifying and retrieving metadata from archive.org items.
"""

# Copyright (C) 2012-2024 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import argparse
import csv
import sys
from collections import defaultdict
from copy import copy
from typing import Mapping

from requests import Response

from internetarchive import item
from internetarchive.cli.cli_utils import MetadataAction, QueryStringAction
from internetarchive.exceptions import ItemLocateError
from internetarchive.utils import json


def setup(subparsers):
    """
    Setup args for metadata command.

    Args:
        subparsers: subparser object passed from ia.py
    """
    parser = subparsers.add_parser("metadata",
                                   aliases=["md"],
                                   help="Retrieve and modify archive.org item metadata")

    parser.add_argument("identifier",
                        nargs="?",
                        type=str,
                        help="Identifier for the upload")

    # Mutually exclusive group for metadata modification options
    modify_group = parser.add_mutually_exclusive_group()
    modify_group.add_argument("-m", "--modify",
                              nargs="+",
                              action=MetadataAction,
                              metavar="KEY:VALUE",
                              help="Modify the metadata of an item")
    modify_group.add_argument("-r", "--remove",
                              nargs="+",
                              action=MetadataAction,
                              metavar="KEY:VALUE",
                              help="Remove KEY:VALUE from a metadata element")
    modify_group.add_argument("-a", "--append",
                              nargs="+",
                              action=MetadataAction,
                              metavar="KEY:VALUE",
                              help="Append a string to a metadata element")
    modify_group.add_argument("-A", "--append-list",
                              nargs="+",
                              action=MetadataAction,
                              metavar="KEY:VALUE",
                              help="Append a field to a metadata element")
    modify_group.add_argument("-i", "--insert",
                              nargs="+",
                              action=MetadataAction,
                              metavar="KEY:VALUE",
                              help=("Insert a value into a multi-value field given "
                                    "an index (e.g. `--insert=collection[0]:foo`)"))

    # Additional options
    parser.add_argument("-E", "--expect",
                        nargs="+",
                        action=MetadataAction,
                        metavar="KEY:VALUE",
                        help=("Test an expectation server-side before applying patch "
                              "to item metadata"))
    parser.add_argument("-H", "--header",
                        nargs="+",
                        action=QueryStringAction,
                        metavar="KEY:VALUE",
                        help="S3 HTTP headers to send with your request")
    parser.add_argument("-t", "--target",
                        metavar="target",
                        default="metadata",
                        help="The metadata target to modify")
    parser.add_argument("-s", "--spreadsheet",
                        metavar="metadata.csv",
                        help="Modify metadata in bulk using a spreadsheet as input")
    parser.add_argument("-e", "--exists",
                        action="store_true",
                        help="Check if an item exists")
    parser.add_argument("-F", "--formats",
                        action="store_true",
                        help="Return the file-formats the given item contains")
    parser.add_argument("-p", "--priority",
                        metavar="priority",
                        help="Set the task priority")
    parser.add_argument("--timeout",
                        metavar="value",
                        help="Set a timeout for metadata writes")
    parser.add_argument("-R", "--reduced-priority",
                        action="store_true",
                        help="Submit task at a reduced priority.")

    parser.set_defaults(func=lambda args: main(args, parser))


def modify_metadata(item: item.Item,
                    metadata: Mapping,
                    args: argparse.Namespace,
                    parser: argparse.ArgumentParser) -> Response:
    """
    Modify metadata helper function.
    """
    append = bool(args.append)
    append_list = bool(args.append_list)
    insert = bool(args.insert)
    try:
        r = item.modify_metadata(metadata, target=args.target, append=append,
                                 expect=args.expect, priority=args.priority,
                                 append_list=append_list, headers=args.header,
                                 insert=insert, reduced_priority=args.reduced_priority,
                                 timeout=args.timeout)
        assert isinstance(r, Response)  # mypy: modify_metadata() -> Request | Response
    except ItemLocateError as exc:
        print(f"{item.identifier} - error: {exc}", file=sys.stderr)
        sys.exit(1)
    if not r.json()["success"]:
        error_msg = r.json()["error"]
        etype = "warning" if "no changes" in r.text else "error"
        print(f"{item.identifier} - {etype} ({r.status_code}): {error_msg}", file=sys.stderr)
        return r
    print(f"{item.identifier} - success: {r.json()['log']}", file=sys.stderr)
    return r


def remove_metadata(item: item.Item,
                    metadata: Mapping,
                    args: argparse.Namespace,
                    parser: argparse.ArgumentParser) -> Response:
    """
    Remove metadata helper function.
    """
    md: dict[str, list | str] = defaultdict(list)
    for key in metadata:
        src_md = {}
        if args.target.startswith("files/"):
            for f in item.get_files():
                if f.name == "/".join(args.target.split("/")[1:]):
                    src_md = f.__dict__.get(key, {})
                    break
        else:
            src_md = copy(item.metadata.get(key, {}))
        if not src_md:
            continue

        if key == "collection":
            _col = copy(metadata[key])
            _src_md = copy(src_md)
            if not isinstance(_col, list):
                _col = [_col]
            if not isinstance(_src_md, list):
                _src_md = [_src_md]  # type: ignore
            for c in _col:
                if c not in _src_md:
                    r = item.remove_from_simplelist(c, "holdings")
                    j = r.json()
                    if j.get("success"):
                        print(f"{item.identifier} - success: {item.identifier} no longer in {c}",
                              file=sys.stderr)
                        sys.exit(0)
                    elif j.get("error", "").startswith("no row to delete for"):
                        print(f"{item.identifier} - success: {item.identifier} no longer in {c}",
                              file=sys.stderr)
                        sys.exit(0)
                    else:
                        print(f"{item.identifier} - error: {j.get('error')}", file=sys.stderr)
                        sys.exit(1)

        if not isinstance(src_md, list):
            if key == "subject":
                if isinstance(src_md, str):
                    src_md = src_md.split(";")
            elif key == "collection":
                print(f"{item.identifier} - error: all collections would be removed, "
                      "not submitting task.", file=sys.stderr)
                sys.exit(1)

            if src_md == metadata[key]:
                md[key] = "REMOVE_TAG"
                continue

        for x in src_md:
            if isinstance(metadata[key], list):
                if x not in metadata[key]:
                    md[key].append(x)  # type: ignore
            else:
                if x != metadata[key]:
                    md[key].append(x)  # type: ignore

        if len(md[key]) == len(src_md):
            del md[key]

    if md.get("collection") == []:
        print(f"{item.identifier} - error: all collections would be removed, not submitting task.",
              file=sys.stderr)
        sys.exit(1)
    elif not md:
        print(f"{item.identifier} - warning: nothing needed to be removed.", file=sys.stderr)
        sys.exit(0)

    r = modify_metadata(item, md, args, parser)
    return r


def main(args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
    """
    Main entry point for 'ia metadata'.
    """
    formats = set()
    responses: list[bool | Response] = []

    item = args.session.get_item(args.identifier)

    # Check existence of item.
    if args.exists:
        if item.exists:
            responses.append(True)
            print(f"{args.identifier} exists", file=sys.stderr)
        else:
            responses.append(False)
            print(f"{args.identifier} does not exist", file=sys.stderr)
        if all(r is True for r in responses):
            sys.exit(0)
        else:
            sys.exit(1)

    # Modify metadata.
    elif (args.modify or args.append or args.append_list
          or args.remove or args.insert):
        # TODO: Find a better way to handle this.
        if args.modify:
            metadata = args.modify
        elif args.append:
            metadata = args.append
        elif args.append_list:
            metadata = args.append_list
        elif args.insert:
            metadata = args.insert
        if args.remove:
            metadata = args.remove

        if args.remove:
            responses.append(remove_metadata(item, metadata, args, parser))
        else:
            responses.append(modify_metadata(item, metadata, args, parser))
        if all(r.status_code == 200 for r in responses):  # type: ignore
            sys.exit(0)
        else:
            for r in responses:
                assert isinstance(r, Response)
                if r.status_code == 200:
                    continue
                # We still want to exit 0 if the non-200 is a
                # "no changes to xml" error.
                elif "no changes" in r.text:
                    continue
                else:
                    sys.exit(1)

    # Get metadata.
    elif args.formats:
        for f in item.get_files():
            formats.add(f.format)
        print("\n".join(formats))

    # Edit metadata for items in bulk, using a spreadsheet as input.
    elif args.spreadsheet:
        if not args.priority:
            args.priority = -5
        with open(args.spreadsheet, newline="", encoding="utf-8-sig") as csvfp:
            spreadsheet = csv.DictReader(csvfp)
            responses = []
            for row in spreadsheet:
                if not row["identifier"]:
                    continue
                item = args.session.get_item(row["identifier"])
                if row.get("file"):
                    del row["file"]
                metadata = {k.lower(): v for k, v in row.items() if v}
                responses.append(modify_metadata(item, metadata, args, parser))

            if all(r.status_code == 200 for r in responses):  # type: ignore
                sys.exit(0)
            else:
                for r in responses:
                    assert isinstance(r, Response)
                    if r.status_code == 200:
                        continue
                    # We still want to exit 0 if the non-200 is a
                    # "no changes to xml" error.
                    elif "no changes" in r.text:
                        continue
                    else:
                        sys.exit(1)

    # Dump JSON to stdout.
    else:
        metadata_str = json.dumps(item.item_metadata)
        print(metadata_str)



"""
ia_flag.py

'ia' subcommand for managing flags on archive.org.
"""

# Copyright (C) 2012-2025 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import argparse


def setup(subparsers):
    """Set up argument parser for the 'flag' subcommand.

    Args:
        subparsers: argparse subparsers object from main CLI
    """
    parser = subparsers.add_parser(
        "flag",
        aliases=["fl"],
        help="Manage flags",
    )
    parser.add_argument(
        "identifier",
        nargs="?",
        type=str,
        help="Identifier for the upload",
    )
    parser.add_argument(
        "-u",
        "--user",
        type=str,
        help="User associated with the flag",
    )

    group = parser.add_argument_group("Add flag operations")
    group.add_argument(
        "-a",
        "--add-flag",
        metavar="CATEGORY",
        type=str,
        help="Add a flag to the item",
    )

    group = parser.add_argument_group("Delete flag operations")
    group.add_argument(
        "-d",
        "--delete-flag",
        metavar="CATEGORY",
        type=str,
        help="Delete a flag from the item",
    )

    parser.set_defaults(func=lambda args: main(args, parser))

def main(args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
    """Handle flag subcommand execution.

    Args:
        args: Parsed command-line arguments
        parser: Argument parser for error handling
    """
    item = args.session.get_item(args.identifier)
    if args.user:
        flag_user = args.user
    else:
        flag_user = args.session.config.get("general", {}).get("screenname")
    if not flag_user.startswith('@'):
        flag_user = f"@{flag_user}"
    if args.add_flag:
        r = item.add_flag(args.add_flag, flag_user)
        j = r.json()
        if j.get("status") == "success":
            print(f"success: added '{args.add_flag}' flag by {flag_user} to {args.identifier}")
        else:
            print(f"error: {item.identifier} - {r.text}")

    elif args.delete_flag:
        r = item.delete_flag(args.delete_flag, flag_user)
        j = r.json()
        if j.get("status") == "success":
            print(f"success: deleted '{args.delete_flag}' flag by {flag_user} from {args.identifier}")
        else:
            print(f"error: {item.identifier} - {r.text}")

    else:
        r = item.get_flags()
        print(r.text)



from copy import deepcopy

from tests.conftest import IaRequestsMock, ia_call

NASA_FILES = {
    'NASAarchiveLogo.jpg',
    'globe_west_540.jpg',
    'nasa_reviews.xml',
    'nasa_meta.xml',
    'nasa_archive.torrent',
    'nasa_files.xml'
}


def test_ia_list(capsys, nasa_mocker):
    ia_call(['ia', 'list', 'nasa'])
    out, err = capsys.readouterr()
    assert {l for l in out.split('\n') if l} == NASA_FILES


def test_ia_list_verbose(capsys, nasa_mocker):
    ia_call(['ia', 'list', '--verbose', 'nasa'])

    out, err = capsys.readouterr()
    _nasa_files = deepcopy(NASA_FILES)
    _nasa_files.add('name')
    assert {l for l in out.split('\n') if l} == _nasa_files


def test_ia_list_all(capsys, nasa_mocker):
    ia_call(['ia', 'list', '--all', 'nasa'])

    out, err = capsys.readouterr()
    out = [l for l in out.split('\n') if l]
    assert len(out) == 6
    assert all(len(f.split('\t')) == 9 for f in out)
    assert all(f.split('\t')[0] in NASA_FILES for f in out)


def test_ia_list_location(capsys, nasa_mocker):
    ia_call(['ia', 'list', '--location', '--glob', '*meta.xml', 'nasa'])
    out, err = capsys.readouterr()
    assert out == 'https://archive.org/download/nasa/nasa_meta.xml\n'


def test_ia_list_columns(capsys):
    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        ia_call(['ia', 'list', '--columns', 'name,md5', '--glob', '*meta.xml', 'nasa'])

    out, err = capsys.readouterr()
    assert out == 'nasa_meta.xml\t0e339f4a29a8bc42303813cbec9243e5\n'

    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        ia_call(['ia', 'list', '--columns', 'md5', '--glob', '*meta.xml', 'nasa'])

    out, err = capsys.readouterr()
    assert out == '0e339f4a29a8bc42303813cbec9243e5\n'


def test_ia_list_glob(capsys, nasa_mocker):
    ia_call(['ia', 'list', '--glob', '*torrent', 'nasa'])
    out, err = capsys.readouterr()
    assert out == 'nasa_archive.torrent\n'


def test_ia_list_format(capsys, nasa_mocker):
    ia_call(['ia', 'list', '--format', 'Metadata', 'nasa'])

    out, err = capsys.readouterr()
    expected_output = {
        'nasa_reviews.xml',
        'nasa_files.xml',
        'nasa_meta.xml',
    }
    assert {f for f in out.split('\n') if f} == expected_output


def test_ia_list_non_existing(capsys):
    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa', body='{}')
        ia_call(['ia', 'list', 'nasa'], expected_exit_code=1)

    out, err = capsys.readouterr()
    assert out == ''



import contextlib
import os
import tempfile
from unittest import mock

import pytest
import requests.adapters
import responses

import internetarchive.config
import internetarchive.session
from internetarchive.exceptions import AuthenticationError


@responses.activate
def test_get_auth_config():
    test_body = """{
        "success": true,
        "values": {
            "cookies": {
                "logged-in-sig": "foo-sig",
                "logged-in-user": "foo%40example.com"
            },
            "email": "foo@example.com",
            "itemname": "@jakej",
            "s3": {
                "access": "Ac3ssK3y",
                "secret": "S3cretK3y"
            },
            "screenname":"jakej"
        },
        "version": 1}"""
    responses.add(responses.POST, 'https://archive.org/services/xauthn/',
                  body=test_body)
    r = internetarchive.config.get_auth_config('test@example.com', 'password1')
    assert r['s3']['access'] == 'Ac3ssK3y'
    assert r['s3']['secret'] == 'S3cretK3y'
    assert r['cookies']['logged-in-user'] == 'foo%40example.com'
    assert r['cookies']['logged-in-sig'] == 'foo-sig'


@responses.activate
def test_get_auth_config_auth_fail():
    # No logged-in-sig cookie set raises AuthenticationError.
    responses.add(responses.POST, 'https://archive.org/services/xauthn/',
                  body='{"error": "failed"}')
    try:
        r = internetarchive.config.get_auth_config('test@example.com', 'password1')
    except AuthenticationError as exc:
        return
        assert str(exc) == ('Authentication failed. Please check your credentials '
                            'and try again.')


def test_get_config():
    config = internetarchive.config.get_config()
    assert isinstance(config, dict)


def test_get_config_with_config_file(tmpdir):
    test_conf = ('[s3]\n'
                 'access = test-access\n'
                 'secret = test-secret\n'
                 '[cookies]\n'
                 'logged-in-sig = test-sig\n'
                 'logged-in-user = test@archive.org\n')

    tmpdir.chdir()
    with open('ia_test.ini', 'w') as fp:
        fp.write(test_conf)

    config = internetarchive.config.get_config(config_file='ia_test.ini',
                                               config={'custom': 'test'})
    assert config['cookies']['logged-in-sig'] == 'test-sig'
    assert config['cookies']['logged-in-user'] == 'test@archive.org'
    assert config['s3']['access'] == 'test-access'
    assert config['s3']['secret'] == 'test-secret'
    assert config['custom'] == 'test'


def test_get_config_no_config_file():
    os.environ['HOME'] = ''
    config = internetarchive.config.get_config()
    assert config == {}


def test_get_config_with_config():
    test_conf = {
        's3': {
            'access': 'custom-access',
            'secret': 'custom-secret',
        },
        'cookies': {
            'logged-in-user': 'test@archive.org',
            'logged-in-sig': 'test-sig',
        },
    }

    os.environ['HOME'] = ''
    config = internetarchive.config.get_config(config=test_conf)
    assert config['cookies']['logged-in-sig'] == 'test-sig'
    assert config['cookies']['logged-in-user'] == 'test@archive.org'
    assert config['s3']['access'] == 'custom-access'
    assert config['s3']['secret'] == 'custom-secret'


def test_get_config_home_not_set():
    os.environ['HOME'] = '/none'
    config = internetarchive.config.get_config()
    assert isinstance(config, dict)


def test_get_config_home_not_set_with_config():
    test_conf = {
        's3': {
            'access': 'no-home-access',
            'secret': 'no-home-secret',
        },
    }
    os.environ['HOME'] = '/none'
    config = internetarchive.config.get_config(config=test_conf)
    assert isinstance(config, dict)
    assert config['s3']['access'] == 'no-home-access'
    assert config['s3']['secret'] == 'no-home-secret'


def test_get_config_config_and_config_file(tmpdir):
    test_conf = ('[s3]\n'
                 'access = test-access\n'
                 'secret = test-secret\n'
                 '[cookies]\n'
                 'logged-in-sig = test-sig\n'
                 'logged-in-user = test@archive.org\n')

    tmpdir.chdir()

    with open('ia_test.ini', 'w') as fp:
        fp.write(test_conf)

    test_conf = {
        's3': {
            'access': 'custom-access',
            'secret': 'custom-secret',
        },
        'cookies': {
            'logged-in-user': 'test@archive.org',
            'logged-in-sig': 'test-sig',
        },
    }
    del test_conf['s3']['access']
    config = internetarchive.config.get_config(config_file='ia_test.ini',
                                               config=test_conf)
    assert config['cookies']['logged-in-sig'] == 'test-sig'
    assert config['cookies']['logged-in-user'] == 'test@archive.org'
    assert config['s3']['access'] == 'test-access'
    assert config['s3']['secret'] == 'custom-secret'


@contextlib.contextmanager
def _environ(**kwargs):
    old_values = {k: os.environ.get(k) for k in kwargs}
    try:
        for k, v in kwargs.items():
            if v is not None:
                os.environ[k] = v
            else:
                del os.environ[k]
        yield
    finally:
        for k, v in old_values.items():
            if v is not None:
                os.environ[k] = v
            else:
                del os.environ[k]


def _test_parse_config_file(
        expected_result,
        config_file_contents='',
        config_file_paths=None,
        home=None,
        xdg_config_home=None,
        config_file_param=None):
    # expected_result: (config_file_path, is_xdg); config isn't compared.
    # config_file_contents: str
    # config_file_paths: list of filenames to write config_file_contents to
    # home: str, override HOME env var; default: path of the temporary dir
    # xdg_config_home: str, set XDG_CONFIG_HOME
    # config_file_param: str, filename to pass to parse_config_file
    # All paths starting with '$TMPTESTDIR/' get evaluated relative to the temp dir.

    if not config_file_paths:
        config_file_paths = []

    with tempfile.TemporaryDirectory() as tmp_test_dir:
        def _replace_path(s):
            if s and s.startswith('$TMPTESTDIR/'):
                return os.path.join(tmp_test_dir, s.split('/', 1)[1])
            return s

        expected_result = (_replace_path(expected_result[0]), expected_result[1])
        config_file_paths = [_replace_path(x) for x in config_file_paths]
        home = _replace_path(home)
        xdg_config_home = _replace_path(xdg_config_home)
        config_file_param = _replace_path(config_file_param)

        for p in config_file_paths:
            os.makedirs(os.path.dirname(p), exist_ok=True)
            with open(p, 'w') as fp:
                fp.write(config_file_contents)

        if home is None:
            home = tmp_test_dir
        env = {'HOME': home}
        if xdg_config_home is not None:
            env['XDG_CONFIG_HOME'] = xdg_config_home
        with _environ(**env):
            config_file_path, is_xdg, config = internetarchive.config.parse_config_file(
                config_file=config_file_param)

    assert (config_file_path, is_xdg) == expected_result[0:2]


def test_parse_config_file_blank():
    _test_parse_config_file(
        expected_result=('$TMPTESTDIR/.config/internetarchive/ia.ini', True)
    )


def test_parse_config_file_existing_config_ia():
    _test_parse_config_file(
        expected_result=('$TMPTESTDIR/.config/ia.ini', False),
        config_file_paths=['$TMPTESTDIR/.config/ia.ini'],
    )


def test_parse_config_file_existing_dotia():
    _test_parse_config_file(
        expected_result=('$TMPTESTDIR/.ia', False),
        config_file_paths=['$TMPTESTDIR/.ia'],
    )


def test_parse_config_file_existing_config_ia_and_dotia():
    _test_parse_config_file(
        expected_result=('$TMPTESTDIR/.config/ia.ini', False),
        config_file_paths=['$TMPTESTDIR/.config/ia.ini', '$TMPTESTDIR/.ia'],
    )


def test_parse_config_file_existing_all():
    _test_parse_config_file(
        expected_result=('$TMPTESTDIR/.config/internetarchive/ia.ini', True),
        config_file_paths=[
            '$TMPTESTDIR/.config/internetarchive/ia.ini',
            '$TMPTESTDIR/.config/ia.ini',
            '$TMPTESTDIR/.ia'
        ],
    )


def test_parse_config_file_custom_xdg():
    _test_parse_config_file(
        expected_result=('$TMPTESTDIR/.xdg/internetarchive/ia.ini', True),
        xdg_config_home='$TMPTESTDIR/.xdg',
    )


def test_parse_config_file_empty_xdg():
    # Empty XDG_CONFIG_HOME should be treated as if not set, i.e. default
    _test_parse_config_file(
        expected_result=('$TMPTESTDIR/.config/internetarchive/ia.ini', True),
        xdg_config_home='',
    )


def test_parse_config_file_relative_xdg():
    # Relative XDG_CONFIG_HOME is invalid and should be ignored, i.e. default ~/.config used instead
    _test_parse_config_file(
        expected_result=('$TMPTESTDIR/.config/internetarchive/ia.ini', True),
        xdg_config_home='relative/.config',
    )


def test_parse_config_file_direct_path_overrides_existing_files():
    _test_parse_config_file(
        expected_result=('/path/to/ia.ini', False),
        config_file_paths=[
            '$TMPTESTDIR/.config/internetarchive/ia.ini',
            '$TMPTESTDIR/.config/ia.ini',
            '$TMPTESTDIR/.ia'
        ],
        config_file_param='/path/to/ia.ini',
    )


def test_parse_config_file_with_environment_variable():
    with _environ(IA_CONFIG_FILE='/inexistent.ia.ini'):
        _test_parse_config_file(
            expected_result=('/inexistent.ia.ini', False),
        )


def test_parse_config_file_with_environment_variable_and_parameter():
    with _environ(IA_CONFIG_FILE='/inexistent.ia.ini'):
        _test_parse_config_file(
            expected_result=('/inexistent.other.ia.ini', False),
            config_file_param='/inexistent.other.ia.ini',
        )


def _test_write_config_file(
        expected_config_file,
        expected_modes,
        dirs=None,
        create_expected_file=False,
        config_file_param=None):
    # expected_config_file: str
    # expected_modes: list of (path, mode) tuples
    # dirs: list of str, directories to create before running write_config_file
    # create_expected_file: bool, create the expected_config_file if True
    # config_file_param: str, filename to pass to write_config_file
    # Both dirs and the config file are created with mode 777 (minus umask).
    # All paths are evaluated relative to a temporary HOME.
    # Mode comparison accounts for the umask; expected_modes does not need to care about it.

    with tempfile.TemporaryDirectory() as temp_home_dir:
        expected_config_file = os.path.join(temp_home_dir, expected_config_file)
        if dirs:
            dirs = [os.path.join(temp_home_dir, d) for d in dirs]
        expected_modes = [(os.path.join(temp_home_dir, p), m) for p, m in expected_modes]
        if config_file_param:
            config_file_param = os.path.join(temp_home_dir, config_file_param)
        with _environ(HOME=temp_home_dir):
            # Need to account for the umask in the expected_modes comparisons.
            # The umask can't just be retrieved, so set and then restore previous value.
            umask = os.umask(0)
            os.umask(umask)
            if dirs:
                for d in dirs:
                    os.mkdir(d)
            if create_expected_file:
                with open(expected_config_file, 'w') as fp:
                    os.chmod(expected_config_file, 0o777)
            config_file = internetarchive.config.write_config_file({}, config_file_param)
            assert config_file == expected_config_file
            assert os.path.isfile(config_file)
            for path, mode in expected_modes:
                actual_mode = os.stat(path).st_mode & 0o777
                assert actual_mode == mode & ~umask


def test_write_config_file_blank():
    """Test that a blank HOME is populated with expected dirs and modes."""
    _test_write_config_file(
        expected_config_file='.config/internetarchive/ia.ini',
        expected_modes=[
            ('.config/internetarchive/ia.ini', 0o600),
            ('.config/internetarchive', 0o700),
            ('.config', 0o700),
        ],
    )


def test_write_config_file_config_existing():
    """Test that .config's permissions remain but ia gets created correctly."""
    _test_write_config_file(
        dirs=['.config'],
        expected_config_file='.config/internetarchive/ia.ini',
        expected_modes=[
            ('.config/internetarchive/ia.ini', 0o600),
            ('.config/internetarchive', 0o700),
            ('.config', 0o777),
        ],
    )


def test_write_config_file_config_internetarchive_existing():
    """Test that directory permissions are left as is"""
    _test_write_config_file(
        dirs=['.config', '.config/internetarchive'],
        expected_config_file='.config/internetarchive/ia.ini',
        expected_modes=[
            ('.config/internetarchive/ia.ini', 0o600),
            ('.config/internetarchive', 0o777),
            ('.config', 0o777),
        ],
    )


def test_write_config_file_existing_file():
    """Test that the permissions of the file are forced to 600"""
    _test_write_config_file(
        dirs=['.config', '.config/internetarchive'],
        expected_config_file='.config/internetarchive/ia.ini',
        create_expected_file=True,
        expected_modes=[
            ('.config/internetarchive/ia.ini', 0o600),
            ('.config/internetarchive', 0o777),
            ('.config', 0o777),
        ],
    )


def test_write_config_file_existing_other_file():
    """Test that the permissions of the file are forced to 600 even outside XDG"""
    _test_write_config_file(
        dirs=['foo'],
        expected_config_file='foo/ia.ini',
        create_expected_file=True,
        config_file_param='foo/ia.ini',
        expected_modes=[
            ('foo/ia.ini', 0o600),
            ('foo', 0o777),
        ],
    )


def test_write_config_file_custom_path_existing():
    """Test the creation of a config file at a custom location"""
    _test_write_config_file(
        dirs=['foo'],
        expected_config_file='foo/ia.ini',
        config_file_param='foo/ia.ini',
        expected_modes=[
            ('foo/ia.ini', 0o600),
            ('foo', 0o777),
        ],
    )


def test_write_config_file_custom_path_not_existing():
    """Ensure that an exception is thrown if the custom path dir doesn't exist"""
    with tempfile.TemporaryDirectory() as temp_home_dir:
        with _environ(HOME=temp_home_dir):
            config_file = os.path.join(temp_home_dir, 'foo/ia.ini')
            with pytest.raises(IOError):
                internetarchive.config.write_config_file({}, config_file)



#!/usr/bin/env python
"""
ia.py

The internetarchive module is a Python/CLI interface to Archive.org.
"""

# Copyright (C) 2012-2024 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import signal
import sys

from internetarchive import __version__, get_session
from internetarchive.cli import (
    ia_account,
    ia_configure,
    ia_copy,
    ia_delete,
    ia_download,
    ia_flag,
    ia_list,
    ia_metadata,
    ia_move,
    ia_reviews,
    ia_search,
    ia_simplelists,
    ia_tasks,
    ia_upload,
)
from internetarchive.cli.cli_utils import exit_on_signal

# Handle broken pipe
try:
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
except AttributeError:
    # Non-unix support
    pass

# Handle <Ctrl-C>
signal.signal(signal.SIGINT, exit_on_signal)


def validate_config_path(path):
    """
    Validate the path to the configuration file.

    Returns:
        str: Validated path to the configuration file.
    """
    if "configure" not in sys.argv:  # Support for adding config to specific file
        file_check = argparse.FileType("r")
        file_check(path)
    return path


def main():
    """
    Main entry point for the CLI.
    """
    parser = argparse.ArgumentParser(
            description="A command line interface to Archive.org.",
            epilog=("Documentation for 'ia' is available at:\n\n\t"
                    "https://archive.org/developers/internetarchive/cli.html\n\n"
                    "See 'ia {command} --help' for help on a specific command."),
            formatter_class=argparse.RawTextHelpFormatter)  # support for \n in epilog

    parser.add_argument("-v", "--version",
                        action="version",
                        version=__version__)
    parser.add_argument("-c", "--config-file",
                        action="store",
                        type=validate_config_path,
                        metavar="FILE",
                        help="path to configuration file")
    parser.add_argument("-l", "--log",
                        action="store_true",
                        default=False,
                        help="enable logging")
    parser.add_argument("-d", "--debug",
                        action="store_true",
                        help="enable debugging")
    parser.add_argument("-i", "--insecure",
                        action="store_true",
                        help="allow insecure connections")
    parser.add_argument("-H", "--host",
                        action="store",
                        help=("host to connect to "
                              "(doesn't work for requests made to s3.us.archive.org)"))

    subparsers = parser.add_subparsers(title="commands",
                                       dest="command",
                                       metavar="{command}")

    # Add subcommand parsers
    ia_account.setup(subparsers)
    ia_configure.setup(subparsers)
    ia_copy.setup(subparsers)
    ia_delete.setup(subparsers)
    ia_download.setup(subparsers)
    ia_flag.setup(subparsers)
    ia_list.setup(subparsers)
    ia_metadata.setup(subparsers)
    ia_move.setup(subparsers)
    ia_reviews.setup(subparsers)
    ia_search.setup(subparsers)
    ia_simplelists.setup(subparsers)
    ia_tasks.setup(subparsers)
    ia_upload.setup(subparsers)

    # Suppress help for alias subcommands
    args = parser.parse_args()

    config: dict[str, dict] = {}
    if args.log:
        config["logging"] = {"level": "INFO"}
    elif args.debug:
        config["logging"] = {"level": "DEBUG"}

    if args.insecure:
        config["general"] = {"secure": False}
    if args.host:
        if config.get("general"):
            config["general"]["host"] = args.host
        else:
            config["general"] = {"host": args.host}

    args.session = get_session(config_file=args.config_file,
                               config=config,
                               debug=args.debug)

    # Check if any arguments were provided
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args.func(args)


if __name__ == "__main__":
    main()



#
# The internetarchive module is a Python/CLI interface to Archive.org.
#
# Copyright (C) 2012-2024 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
internetarchive.utils
~~~~~~~~~~~~~~~~~~~~~

This module provides utility functions for the internetarchive library.

:copyright: (C) 2012-2024 by Internet Archive.
:license: AGPL 3, see LICENSE for more details.
"""
from __future__ import annotations

import hashlib
import os
import re
import sys
from collections.abc import Mapping
from typing import Iterable
from xml.dom.minidom import parseString

# Make preferred JSON package available via `from internetarchive.utils import json`
try:
    import ujson as json

    # ujson lacks a JSONDecodeError: https://github.com/ultrajson/ultrajson/issues/497
    JSONDecodeError = ValueError
except ImportError:
    import json  # type: ignore
    JSONDecodeError = json.JSONDecodeError  # type: ignore


def deep_update(d: dict, u: Mapping) -> dict:
    for k, v in u.items():
        if isinstance(v, Mapping):
            r = deep_update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


class InvalidIdentifierException(Exception):
    pass


def validate_s3_identifier(string: str) -> bool:
    legal_chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789._-'
    # periods, underscores, and dashes are legal, but may not be the first
    # character!
    if any(string.startswith(c) is True for c in ['.', '_', '-']):
        raise InvalidIdentifierException('Identifier cannot begin with periods ".", underscores '
                                        '"_", or dashes "-".')

    if len(string) > 100 or len(string) < 3:
        raise InvalidIdentifierException('Identifier should be between 3 and 80 characters in '
                                        'length.')

    # Support for uploading to user items, e.g. first character can be `@`.
    if string.startswith('@'):
        string = string[1:]

    if any(c not in legal_chars for c in string):
        raise InvalidIdentifierException('Identifier can only contain alphanumeric characters, '
                                        'periods ".", underscores "_", or dashes "-". However, '
                                        'identifier cannot begin with periods, underscores, or '
                                        'dashes.')

    return True


def needs_quote(s: str) -> bool:
    try:
        s.encode('ascii')
    except (UnicodeDecodeError, UnicodeEncodeError):
        return True
    return re.search(r'\s', s) is not None


def norm_filepath(fp: bytes | str) -> str:
    if isinstance(fp, bytes):
        fp = fp.decode('utf-8')
    fp = fp.replace(os.path.sep, '/')
    if not fp.startswith('/'):
        fp = f'/{fp}'
    return fp


def get_md5(file_object) -> str:
    m = hashlib.md5()
    while True:
        data = file_object.read(8192)
        if not data:
            break
        m.update(data)
    file_object.seek(0, os.SEEK_SET)
    return m.hexdigest()


def chunk_generator(fp, chunk_size: int):
    while True:
        chunk = fp.read(chunk_size)
        if not chunk:
            break
        yield chunk


def suppress_keyboard_interrupt_message() -> None:
    """Register a new excepthook to suppress KeyboardInterrupt
    exception messages, and exit with status code 130.

    """
    old_excepthook = sys.excepthook

    def new_hook(type, value, traceback):
        if type is KeyboardInterrupt:
            sys.exit(130)
        old_excepthook(type, value, traceback)

    sys.excepthook = new_hook


class IterableToFileAdapter:
    def __init__(self, iterable, size: int, pre_encode: bool = False):
        self.iterator = iter(iterable)
        self.length = size
        # pre_encode is needed because http doesn't know that it
        # needs to encode a TextIO object when it's wrapped
        # in the Iterator from tqdm.
        # So, this FileAdapter provides pre-encoded output
        self.pre_encode = pre_encode

    def read(self, size: int = -1):  # TBD: add buffer for `len(data) > size` case
        if self.pre_encode:
            # this adapter is intended to emulate the encoding that is usually
            # done by the http lib.
            # As of 2022, iso-8859-1 encoding is used to meet the HTTP standard,
            # see in the cpython repo (https://github.com/python/cpython
            # Lib/http/client.py lines 246; 1340; or grep 'iso-8859-1'
            return next(self.iterator, '').encode("iso-8859-1")
        return next(self.iterator, b'')

    def __len__(self) -> int:
        return self.length


class IdentifierListAsItems:
    """This class is a lazily-loaded list of Items, accessible by index or identifier.
    """

    def __init__(self, id_list_or_single_id, session):
        self.ids = (id_list_or_single_id
                    if isinstance(id_list_or_single_id, list)
                    else [id_list_or_single_id])
        self._items = [None] * len(self.ids)
        self.session = session

    def __len__(self) -> int:
        return len(self.ids)

    def __getitem__(self, idx):
        for i in (range(*idx.indices(len(self))) if isinstance(idx, slice) else [idx]):
            if self._items[i] is None:
                self._items[i] = self.session.get_item(self.ids[i])
        return self._items[idx]

    def __getattr__(self, name):
        try:
            return self[self.ids.index(name)]
        except ValueError:
            raise AttributeError

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.ids!r})'


def get_s3_xml_text(xml_str: str) -> str:
    def _get_tag_text(tag_name, xml_obj):
        text = ''
        elements = xml_obj.getElementsByTagName(tag_name)
        for e in elements:
            for node in e.childNodes:
                if node.nodeType == node.TEXT_NODE:
                    text += node.data
        return text

    tag_names = ['Message', 'Resource']
    try:
        p = parseString(xml_str)
        _msg = _get_tag_text('Message', p)
        _resource = _get_tag_text('Resource', p)
        # Avoid weird Resource text that contains PUT method.
        if _resource and "'PUT" not in _resource:
            return f'{_msg} - {_resource.strip()}'
        else:
            return _msg
    except Exception:
        return str(xml_str)


def get_file_size(file_obj) -> int | None:
    if is_filelike_obj(file_obj):
        try:
            file_obj.seek(0, os.SEEK_END)
            size = file_obj.tell()
            # Avoid OverflowError.
            if size > sys.maxsize:
                size = None
            file_obj.seek(0, os.SEEK_SET)
        except OSError:
            size = None
    else:
        st = os.stat(file_obj)
        size = st.st_size
    return size


def iter_directory(directory: str):
    """Given a directory, yield all files recursively as a two-tuple (filepath, s3key)"""
    for path, _dir, files in os.walk(directory):
        for f in files:
            filepath = os.path.join(path, f)
            key = os.path.relpath(filepath, directory)
            yield (filepath, key)


def recursive_file_count_and_size(files, item=None, checksum=False):
    """Given a filepath or list of filepaths, return the total number and size of files.
    If `checksum` is `True`, skip over files whose MD5 hash matches any file in the `item`.
    """
    if not isinstance(files, (list, set)):
        files = [files]
    total_files = 0
    total_size = 0
    if checksum is True:
        md5s = [f.get('md5') for f in item.files]
    else:
        md5s = []
    if isinstance(files, dict):
        # make sure to use local filenames.
        _files = files.values()
    else:
        if isinstance(files[0], tuple):
            _files = dict(files).values()
        else:
            _files = files
    for f in _files:
        try:
            is_dir = os.path.isdir(f)
        except TypeError:
            try:
                f = f[0]
                is_dir = os.path.isdir(f)
            except (AttributeError, TypeError):
                is_dir = False
        if is_dir:
            it = iter_directory(f)
        else:
            it = [(f, None)]
        for x, _ in it:
            if checksum is True:
                try:
                    with open(x, 'rb') as fh:
                        lmd5 = get_md5(fh)
                except TypeError:
                    # Support file-like objects.
                    lmd5 = get_md5(x)
                if lmd5 in md5s:
                    continue
            total_size += get_file_size(x)
            total_files += 1
    return total_files, total_size


def recursive_file_count(*args, **kwargs):
    """Like `recursive_file_count_and_size`, but returns only the file count."""
    total_files, _ = recursive_file_count_and_size(*args, **kwargs)
    return total_files


def is_dir(obj) -> bool:
    """Special is_dir function to handle file-like object cases that
    cannot be stat'd"""
    try:
        return os.path.isdir(obj)
    except TypeError as exc:
        return False


def is_filelike_obj(obj) -> bool:
    """Distinguish file-like from path-like objects"""
    try:
        os.fspath(obj)
    except TypeError:
        return True
    else:
        return False


def reraise_modify(
    caught_exc: Exception,
    append_msg: str,
    prepend: bool = False,
) -> None:
    """Append message to exception while preserving attributes.

    Preserves exception class, and exception traceback.

    Note:
        This function needs to be called inside an except because an exception
        must be active in the current scope.

    Args:
        caught_exc(Exception): The caught exception object
        append_msg(str): The message to append to the caught exception
        prepend(bool): If True prepend the message to args instead of appending

    Returns:
        None

    Side Effects:
        Re-raises the exception with the preserved data / trace but
        modified message
    """
    if not caught_exc.args:
        # If no args, create our own tuple
        arg_list = [append_msg]
    else:
        # Take the last arg
        # If it is a string
        # append your message.
        # Otherwise append it to the
        # arg list(Not as pretty)
        arg_list = list(caught_exc.args[:-1])
        last_arg = caught_exc.args[-1]
        if isinstance(last_arg, str):
            if prepend:
                arg_list.append(append_msg + last_arg)
            else:
                arg_list.append(last_arg + append_msg)
        else:
            arg_list += [last_arg, append_msg]
    caught_exc.args = tuple(arg_list)
    raise  # noqa: PLE0704


def remove_none(obj):
    if isinstance(obj, (list, tuple, set)):
        lst = type(obj)(remove_none(x) for x in obj if x)
        try:
            return [dict(t) for t in {tuple(sorted(d.items())) for d in lst}]
        except (AttributeError, TypeError):
            return lst
    elif isinstance(obj, dict):
        return type(obj)((remove_none(k), remove_none(v))
                         for k, v in obj.items() if k is not None and v is not None)
    else:
        return obj


def delete_items_from_dict(d: dict | list, to_delete):
    """Recursively deletes items from a dict,
    if the item's value(s) is in ``to_delete``.
    """
    if not isinstance(to_delete, list):
        to_delete = [to_delete]
    if isinstance(d, dict):
        for single_to_delete in set(to_delete):
            if single_to_delete in d.values():
                for k, v in d.copy().items():
                    if v == single_to_delete:
                        del d[k]
        for v in d.values():
            delete_items_from_dict(v, to_delete)
    elif isinstance(d, list):
        for i in d:
            delete_items_from_dict(i, to_delete)
    return remove_none(d)


def is_valid_metadata_key(name: str) -> bool:
    # According to the documentation a metadata key
    # has to be a valid XML tag name.
    #
    # The actual allowed tag names (at least as tested with the metadata API),
    # are way more restrictive and only allow ".-A-Za-z_", possibly followed
    # by an index in square brackets e. g. [0].
    # On the other hand the Archive allows tags starting with the string "xml".
    return bool(re.fullmatch(r'[A-Za-z][.\-0-9A-Za-z_]+(?:\[[0-9]+\])?', name))


def merge_dictionaries(
    dict0: dict | None,
    dict1: dict | None,
    keys_to_drop: Iterable | None = None,
) -> dict:
    """Merge two dictionaries.

       Items in `dict0` can optionally be dropped before the merge.

       If equal keys exist in both dictionaries,
       entries in`dict0` are overwritten.

       :param dict0: A base dictionary with the bulk of the items.

       :param dict1: Additional items which overwrite the items in `dict0`.

       :param keys_to_drop: An iterable of keys to drop from `dict0` before the merge.

       :returns: A merged dictionary.
       """
    if dict0 is not None:
        new_dict = dict0.copy()
    else:
        new_dict = {}

    if keys_to_drop is not None:
        for key in keys_to_drop:
            new_dict.pop(key, None)

    # Items from `dict1` take precedence over items from `dict0`.
    if dict1 is not None:
        new_dict.update(dict1)

    return new_dict


def parse_dict_cookies(value: str) -> dict[str, str | None]:
    result: dict[str, str | None] = {}
    for item in value.split(';'):
        item = item.strip()
        if not item:
            continue
        if '=' not in item:
            result[item] = None
            continue
        name, value = item.split('=', 1)
        result[name] = value
    if 'domain' not in result:
        result['domain'] = '.archive.org'
    if 'path' not in result:
        result['path'] = '/'
    return result


def is_valid_email(email):
    # Regular expression pattern for a valid email address
    # Ensures the TLD has at least 2 characters
    pattern = r'^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z]{2,}$'
    return re.match(pattern, email) is not None



import os
import re

import responses
import urllib3

from internetarchive import (
    download,
    get_files,
    get_item,
    get_session,
    modify_metadata,
    search_items,
    upload,
)
from internetarchive.utils import InvalidIdentifierException, json
from tests.conftest import (
    NASA_METADATA_PATH,
    PROTOCOL,
    IaRequestsMock,
    load_file,
    load_test_data_file,
)

TEST_SEARCH_RESPONSE = load_test_data_file('advanced_search_response.json')
TEST_SCRAPE_RESPONSE = load_test_data_file('scrape_response.json')
_j = json.loads(TEST_SCRAPE_RESPONSE)
del _j['cursor']
_j['items'] = [{'identifier': 'nasa'}]
_j['total'] = 1
TEST_SCRAPE_RESPONSE = json.dumps(_j)


def test_get_session_with_config():
    s = get_session(config={'s3': {'access': 'key'}, 'gengeral': {'secure': False}})
    assert s.access_key == 'key'


def test_get_session_with_config_file(tmpdir):
    tmpdir.chdir()
    test_conf = '[s3]\naccess = key2'
    with open('ia_test.ini', 'w') as fh:
        fh.write(test_conf)
    s = get_session(config_file='ia_test.ini')
    assert s.access_key == 'key2'


def test_get_item(nasa_mocker):
    item = get_item('nasa')
    assert item.identifier == 'nasa'


def test_get_item_with_config(nasa_mocker):
    item = get_item('nasa', config={'s3': {'access': 'key'}})
    assert item.session.access_key == 'key'


def test_get_item_with_config_file(tmpdir, nasa_mocker):
    tmpdir.chdir()
    test_conf = '[s3]\naccess = key2'
    with open('ia_test.ini', 'w') as fh:
        fh.write(test_conf)

    item = get_item('nasa', config_file='ia_test.ini')
    assert item.session.access_key == 'key2'


def test_get_item_with_archive_session(nasa_mocker):
    s = get_session(config={'s3': {'access': 'key3'}})
    item = get_item('nasa', archive_session=s)
    assert item.session.access_key == 'key3'


def test_get_item_with_kwargs():
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        item = get_item('nasa', http_adapter_kwargs={'max_retries': 13})
        assert isinstance(item.session.adapters[f'{PROTOCOL}//'].max_retries,
                          urllib3.Retry)

    try:
        get_item('nasa', request_kwargs={'timeout': .0000000000001})
    except Exception as exc:
        assert 'timed out' in str(exc)


def test_get_files():
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        files = get_files('nasa')
        expected_files = {
            'NASAarchiveLogo.jpg',
            'globe_west_540.jpg',
            'nasa_reviews.xml',
            'nasa_meta.xml',
            'nasa_archive.torrent',
            'nasa_files.xml',
        }
        assert {f.name for f in files} == expected_files


def test_get_files_with_get_item_kwargs(tmpdir):
    tmpdir.chdir()
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        s = get_session(config={'s3': {'access': 'key'}})
        files = get_files('nasa', files='nasa_meta.xml', archive_session=s)
        files = list(files)
        assert len(files) == 1
        assert files[0].name == 'nasa_meta.xml'

        files = get_files('nasa',
                          files='nasa_meta.xml',
                          config={'logging': {'level': 'INFO'}})
        files = list(files)
        assert len(files) == 1
        assert files[0].name == 'nasa_meta.xml'

        test_conf = '[s3]\naccess = key2'
        with open('ia_test.ini', 'w') as fh:
            fh.write(test_conf)
        files = get_files('nasa', files='nasa_meta.xml',
                          config_file='ia_test.ini')
        files = list(files)
        assert len(files) == 1
        assert files[0].name == 'nasa_meta.xml'

        files = get_files('nasa',
                          files='nasa_meta.xml',
                          http_adapter_kwargs={'max_retries': 3})
        files = list(files)
        assert len(files) == 1
        assert files[0].name == 'nasa_meta.xml'

        files = get_files('nasa', files='nasa_meta.xml',
                          request_kwargs={'timeout': 4})
        files = list(files)
        assert len(files) == 1
        assert files[0].name == 'nasa_meta.xml'


def test_get_files_non_existing(nasa_mocker):
    files = get_files('nasa', files='none')
    assert list(files) == []


def test_get_files_multiple(nasa_mocker):
    _files = ['nasa_meta.xml', 'nasa_files.xml']
    files = get_files('nasa', files=_files)
    for f in files:
        assert f.name in _files


def test_get_files_formats():
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        files = get_files('nasa', formats='JPEG')
        files = list(files)
        assert len(files) == 1
        assert files[0].name == 'globe_west_540.jpg'

        files = get_files('nasa', formats=['JPEG', 'Collection Header'])
        expected_files = {
            'globe_west_540.jpg',
            'NASAarchiveLogo.jpg',
        }
        assert {f.name for f in files} == expected_files


def test_get_files_glob_pattern():
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        files = get_files('nasa', glob_pattern='*torrent')
        files = list(files)
        assert len(files) == 1
        assert files[0].name == 'nasa_archive.torrent'

        files = get_files('nasa', glob_pattern='*torrent|*jpg')
        expected_files = {
            'globe_west_540.jpg',
            'NASAarchiveLogo.jpg',
            'nasa_archive.torrent',
        }
        assert {f.name for f in files} == expected_files


def test_modify_metadata():
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(responses.GET, f'{PROTOCOL}//archive.org/metadata/nasa',
                 body='{"metadata":{"title":"foo"}}')
        rsps.add(responses.POST, f'{PROTOCOL}//archive.org/metadata/nasa',
                 body=('{"success":true,"task_id":423444944,'
                       '"log":"https://catalogd.archive.org/log/423444944"}'))
        r = modify_metadata('nasa', {'foo': 1})
        assert r.status_code == 200
        assert r.json() == {
            'task_id': 423444944,
            'success': True,
            'log': 'https://catalogd.archive.org/log/423444944'
        }


def test_upload():
    expected_s3_headers = {
        'content-length': '7557',
        'x-archive-queue-derive': '1',
        'x-archive-meta00-scanner': 'uri(Internet%20Archive%20Python%20library',
        'x-archive-size-hint': '7557',
        'x-archive-auto-make-bucket': '1',
        'authorization': 'LOW test_access:test_secret',
    }
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(responses.PUT, re.compile(r'.*s3.us.archive.org/.*'),
                 adding_headers=expected_s3_headers)
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.GET, f'{PROTOCOL}//archive.org/metadata/nasa',
                 body='{}')
        _responses = upload('nasa', NASA_METADATA_PATH,
                            access_key='test_access',
                            secret_key='test_secret')
        for response in _responses:
            req = response.request
            headers = {k.lower(): str(v) for k, v in req.headers.items()}
            scanner_header = '%20'.join(
                response.headers['x-archive-meta00-scanner'].split('%20')[:4])
            headers['x-archive-meta00-scanner'] = scanner_header
            assert 'user-agent' in headers
            del headers['accept']
            del headers['accept-encoding']
            del headers['connection']
            del headers['user-agent']
            assert headers == expected_s3_headers
            assert req.url == f'{PROTOCOL}//s3.us.archive.org/nasa/nasa.json'


def test_upload_validate_identifier():
    try:
        upload('føø', NASA_METADATA_PATH,
              access_key='test_access',
              secret_key='test_secret',
              validate_identifier=True)
        raise AssertionError("Given invalid identifier was not correctly validated.")
    except Exception as exc:
        assert isinstance(exc, InvalidIdentifierException)

    expected_s3_headers = {
        'content-length': '7557',
        'x-archive-queue-derive': '1',
        'x-archive-meta00-scanner': 'uri(Internet%20Archive%20Python%20library',
        'x-archive-size-hint': '7557',
        'x-archive-auto-make-bucket': '1',
        'authorization': 'LOW test_access:test_secret',
    }
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(responses.PUT, re.compile(r'.*s3.us.archive.org/.*'),
                 adding_headers=expected_s3_headers)
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.GET, f'{PROTOCOL}//archive.org/metadata/nasa',
                 body='{}')
        upload('nasa', NASA_METADATA_PATH,
              access_key='test_access',
              secret_key='test_secret',
              validate_identifier=True)
        assert True


def test_download(tmpdir):
    tmpdir.chdir()
    last_mod_header = {"Last-Modified": "Tue, 14 Nov 2023 20:25:48 GMT"}
    with IaRequestsMock() as rsps:
        rsps.add(responses.GET,
                 f'{PROTOCOL}//archive.org/download/nasa/nasa_meta.xml',
                 body='test content',
                 adding_headers=last_mod_header)
        rsps.add_metadata_mock('nasa')
        download('nasa', 'nasa_meta.xml')
        p = os.path.join(str(tmpdir), 'nasa')
        assert len(os.listdir(p)) == 1
        assert load_file('nasa/nasa_meta.xml') == 'test content'


def test_search_items(session):
    url = f'{PROTOCOL}//archive.org/services/search/v1/scrape'
    p1 = {
        'q': 'identifier:nasa',
        'count': '10000',
    }
    p2 = p1.copy()
    p2['total_only'] = 'true'
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(responses.POST, url,
                 body=TEST_SCRAPE_RESPONSE,
                 match=[responses.matchers.query_param_matcher(p1)])
        rsps.add(responses.POST, url,
                 body='{"items":[],"count":0,"total":1}',
                 match=[responses.matchers.query_param_matcher(p2)],
                 content_type='application/json; charset=UTF-8')
        r = search_items('identifier:nasa', archive_session=session)
        expected_results = [{'identifier': 'nasa'}]
        assert r.num_found == 1
        assert iter(r).search == r
        assert len(iter(r)) == 1
        assert len(r.iter_as_results()) == 1
        assert list(r) == expected_results
        assert list(r.iter_as_results()) == expected_results
        assert r.fts == False


def test_search_items_with_fields(session):
    _j = json.loads(TEST_SCRAPE_RESPONSE)
    _j['items'] = [
        {'identifier': 'nasa', 'title': 'NASA Images'}
    ]
    search_response_str = json.dumps(_j)
    url = f'{PROTOCOL}//archive.org/services/search/v1/scrape'
    p1 = {
        'q': 'identifier:nasa',
        'count': '10000',
        'fields': 'identifier,title',
    }
    p2 = {
        'q': 'identifier:nasa',
        'total_only': 'true',
        'count': '10000',
    }
    with IaRequestsMock() as rsps:
        rsps.add(responses.POST, url,
                 match=[responses.matchers.query_param_matcher(p1)],
                 body=search_response_str)
        rsps.add(responses.POST, url,
                 body='{"items":[],"count":0,"total":1}',
                 match=[responses.matchers.query_param_matcher(p2)],
                 content_type='application/json; charset=UTF-8')
        r = search_items('identifier:nasa', fields=['identifier', 'title'],
                         archive_session=session)
        assert list(r) == [{'identifier': 'nasa', 'title': 'NASA Images'}]


def test_search_items_as_items(session):
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(responses.POST,
                 f'{PROTOCOL}//archive.org/services/search/v1/scrape',
                 body=TEST_SCRAPE_RESPONSE)
        rsps.add_metadata_mock('nasa')
        r = search_items('identifier:nasa', archive_session=session)
        assert [x.identifier for x in r.iter_as_items()] == ['nasa']
        assert r.iter_as_items().search == r


def test_search_items_fts(session):
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(responses.POST,
                 f'{PROTOCOL}//be-api.us.archive.org/ia-pub-fts-api',
                 body=TEST_SCRAPE_RESPONSE)
        rsps.add_metadata_mock('nasa')

        r = search_items('nina simone',
                         full_text_search=True,
                         archive_session=session)
        print(r.search_url)
        assert r.fts == True
        assert r.dsl_fts == False
        assert r.query == '!L nina simone'
        assert r.params == {'count': 10000, 'q': '!L nina simone'}

        r = search_items('nina simone',
                         full_text_search=True,
                         dsl_fts=True,
                         archive_session=session)
        assert r.fts == True
        assert r.dsl_fts == True
        assert r.query == 'nina simone'
        assert r.params == {'count': 10000, 'q': 'nina simone'}
        r = search_items('nina simone',
                         dsl_fts=True,
                         archive_session=session)
        assert r.fts == True
        assert r.dsl_fts == True
        assert r.query == 'nina simone'
        assert r.params == {'count': 10000, 'q': 'nina simone'}


def test_page_row_specification(session):
    _j = json.loads(TEST_SEARCH_RESPONSE)
    _j['response']['items'] = [{'identifier': 'nasa'}]
    _j['response']['numFound'] = 1
    _search_r = json.dumps(_j)
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(responses.GET, f'{PROTOCOL}//archive.org/advancedsearch.php',
                 body=_search_r)
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.POST,
                 f'{PROTOCOL}//archive.org/services/search/v1/scrape',
                 body='{"items":[],"count":0,"total":1}',
                 content_type='application/json; charset=UTF-8')
        r = search_items('identifier:nasa', params={'page': '1', 'rows': '1'},
                         archive_session=session)
        assert r.iter_as_items().search == r
        assert len(r.iter_as_items()) == 1



"""
ia_search.py

'ia' subcommand for searching items on archive.org.
"""

# Copyright (C) 2012-2024 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import argparse
import sys
from itertools import chain

from requests.exceptions import ConnectTimeout, ReadTimeout

from internetarchive.cli.cli_utils import FlattenListAction, QueryStringAction
from internetarchive.exceptions import AuthenticationError
from internetarchive.utils import json


def setup(subparsers):
    """
    Setup args for search command.

    Args:
        subparsers: subparser object passed from ia.py
    """
    parser = subparsers.add_parser("search",
                                   aliases=["se"],
                                   help="Search items on archive.org")

    # Positional arguments
    parser.add_argument("query",
                        type=str,
                        help="Search query or queries.")

    # Optional arguments
    parser.add_argument("-p", "--parameters",
                        nargs="+",
                        action=QueryStringAction,
                        metavar="KEY:VALUE",
                        help="Parameters to send with your query.")
    parser.add_argument("-H", "--header",
                        nargs="+",
                        action=QueryStringAction,
                        metavar="KEY:VALUE",
                        help="Add custom headers to your search request.")
    parser.add_argument("-s", "--sort",
                        action="append",
                        help=("Sort search results by specified fields. "
                              "See https://archive.org/advancedsearch.php "
                              "for full list of sort values"
                              " (e.g. --sort 'date desc', --sort 'date asc', etc.)."))
    parser.add_argument("-i", "--itemlist",
                        action="store_true",
                        help="Output identifiers only.")
    parser.add_argument("-f", "--field",
                        nargs="+",
                        action=FlattenListAction,
                        help="Metadata fields to return.")
    parser.add_argument("-n", "--num-found",
                        action="store_true",
                        help="Print the number of results to stdout.")
    parser.add_argument("-F", "--fts",
                        action="store_true",
                        help="Beta support for querying the archive.org full text search API.")
    parser.add_argument("-D", "--dsl-fts",
                        action="store_true",
                        help="Submit --fts query in dsl.")
    parser.add_argument("-t", "--timeout",
                        type=float,
                        default=300,
                        help="Set the timeout in seconds.")


    parser.set_defaults(func=lambda args: main(args, parser))


def prepare_values(value):
    """
    Prepare comma-separated values based on the input value.
    """
    if value:
        return list(chain.from_iterable([x.split(",") for x in value]))
    return None


def perform_search(args, fields, sorts, r_kwargs):
    """
    Perform the search using the provided arguments and request kwargs.
    """
    return args.session.search_items(args.query,  # type: ignore
                                     fields=fields,
                                     sorts=sorts,
                                     params=args.parameters,
                                     full_text_search=args.fts,
                                     dsl_fts=args.dsl_fts,
                                     request_kwargs=r_kwargs)


def handle_search_results(args, search):
    """
    Handle search results based on command-line arguments.
    """
    if args.num_found:
        print(search.num_found)
        sys.exit(0)

    for result in search:
        if args.itemlist:
            if args.fts or args.dsl_fts:
                print("\n".join(result.get("fields", {}).get("identifier")))
            else:
                print(result.get("identifier", ""))
        else:
            print(json.dumps(result))
            if result.get("error"):
                sys.exit(1)


def handle_value_error(exc):
    """
    Handle ValueError exception.
    """
    return f"error: {exc}"


def handle_connect_timeout():
    """
    Handle ConnectTimeout exception.
    """
    return "error: Request timed out. Increase the --timeout and try again."


def handle_read_timeout():
    """
    Handle ReadTimeout exception.
    """
    return "error: The server timed out and failed to return all search results, please try again"


def handle_authentication_error(exc):
    """
    Handle AuthenticationError exception.
    """
    return f"error: {exc}"


def main(args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
    """
    Main entry point for 'ia search'.
    """
    try:
        # Prepare fields and sorts.
        fields = prepare_values(args.field)
        sorts = prepare_values(args.sort)

        # Prepare request kwargs.
        r_kwargs = {
            "headers": args.header,
            "timeout": args.timeout,
        }

        # Perform search.
        search = perform_search(args, fields, sorts, r_kwargs)

        # Handle search results.
        handle_search_results(args, search)

    except ValueError as exc:
        error_message = handle_value_error(exc)
        print(error_message, file=sys.stderr)
        sys.exit(1)

    except ConnectTimeout:
        error_message = handle_connect_timeout()
        print(error_message, file=sys.stderr)
        sys.exit(1)

    except ReadTimeout:
        error_message = handle_read_timeout()
        print(error_message, file=sys.stderr)
        sys.exit(1)

    except AuthenticationError as exc:
        error_message = handle_authentication_error(exc)
        print(error_message, file=sys.stderr)
        sys.exit(1)



#
# The internetarchive module is a Python/CLI interface to Archive.org.
#
# Copyright (C) 2012-2024 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
internetarchive.config
~~~~~~~~~~~~~~~~~~~~~~

:copyright: (C) 2012-2024 by Internet Archive.
:license: AGPL 3, see LICENSE for more details.
"""
from __future__ import annotations

import os
from collections import defaultdict
from configparser import RawConfigParser
from time import sleep
from typing import Mapping

import requests

from internetarchive import auth
from internetarchive.exceptions import AuthenticationError
from internetarchive.utils import deep_update


def get_auth_config(email: str, password: str, host: str = 'archive.org') -> dict:
    u = f'https://{host}/services/xauthn/'
    p = {'op': 'login'}
    d = {'email': email, 'password': password}
    r = requests.post(u, params=p, data=d, timeout=10)
    sleep(2)
    j = r.json()
    if not j.get('success'):
        try:
            msg = j['values']['reason']
        except KeyError:
            msg = j['error']
        if msg == 'account_not_found':
            msg = 'Account not found, check your email and try again.'
        elif msg == 'account_bad_password':
            msg = 'Incorrect password, try again.'
        else:
            msg = f'Authentication failed: {msg}'
        raise AuthenticationError(msg)
    auth_config = {
        's3': {
            'access': j['values']['s3']['access'],
            'secret': j['values']['s3']['secret'],
        },
        'cookies': {
            'logged-in-user': j['values']['cookies']['logged-in-user'],
            'logged-in-sig': j['values']['cookies']['logged-in-sig'],
        },
        'general': {
            'screenname': j['values']['screenname'],
        }
    }
    return auth_config


def write_config_file(auth_config: Mapping, config_file=None):
    config_file, is_xdg, config = parse_config_file(config_file)

    # S3 Keys.
    access = auth_config.get('s3', {}).get('access')
    secret = auth_config.get('s3', {}).get('secret')
    config.set('s3', 'access', access)
    config.set('s3', 'secret', secret)

    # Cookies.
    cookies = auth_config.get('cookies', {})
    config.set('cookies', 'logged-in-user', cookies.get('logged-in-user'))
    config.set('cookies', 'logged-in-sig', cookies.get('logged-in-sig'))

    # General.
    screenname = auth_config.get('general', {}).get('screenname')
    config.set('general', 'screenname', screenname)

    # Create directory if needed.
    config_directory = os.path.dirname(config_file)
    if is_xdg and not os.path.exists(config_directory):
        # os.makedirs does not apply the mode for intermediate directories since Python 3.7.
        # The XDG Base Dir spec requires that the XDG_CONFIG_HOME directory be created with mode 700.
        # is_xdg will be True iff config_file is ${XDG_CONFIG_HOME}/internetarchive/ia.ini.
        # So create grandparent first if necessary then parent to ensure both have the right mode.
        os.makedirs(os.path.dirname(config_directory), mode=0o700, exist_ok=True)
        os.mkdir(config_directory, 0o700)

    # Write config file.
    with open(config_file, 'w') as fh:
        os.chmod(config_file, 0o600)
        config.write(fh)

    return config_file


def parse_config_file(config_file=None):
    config = RawConfigParser()

    is_xdg = False
    if not config_file:
        candidates = []
        if os.environ.get('IA_CONFIG_FILE'):
            candidates.append(os.environ['IA_CONFIG_FILE'])
        xdg_config_home = os.environ.get('XDG_CONFIG_HOME')
        if not xdg_config_home or not os.path.isabs(xdg_config_home):
            # Per the XDG Base Dir specification, this should be $HOME/.config. Unfortunately, $HOME
            # does not exist on all systems. Therefore, we use ~/.config here. On a POSIX-compliant
            # system, where $HOME must always be set, the XDG spec will be followed precisely.
            xdg_config_home = os.path.join(os.path.expanduser('~'), '.config')
        xdg_config_file = os.path.join(xdg_config_home, 'internetarchive', 'ia.ini')
        candidates.append(xdg_config_file)
        candidates.append(os.path.join(os.path.expanduser('~'), '.config', 'ia.ini'))
        candidates.append(os.path.join(os.path.expanduser('~'), '.ia'))
        for candidate in candidates:
            if os.path.isfile(candidate):
                config_file = candidate
                break
        else:
            # None of the candidates exist, default to IA_CONFIG_FILE if set else XDG
            config_file = os.environ.get('IA_CONFIG_FILE', xdg_config_file)
        if config_file == xdg_config_file:
            is_xdg = True
    config.read(config_file)

    if not config.has_section('s3'):
        config.add_section('s3')
        config.set('s3', 'access', None)
        config.set('s3', 'secret', None)
    if not config.has_section('cookies'):
        config.add_section('cookies')
        config.set('cookies', 'logged-in-user', None)
        config.set('cookies', 'logged-in-sig', None)

    if config.has_section('general'):
        for k, _v in config.items('general'):
            if k in ['secure']:
                config.set('general', k, config.getboolean('general', k))
        if not config.get('general', 'screenname'):
            config.set('general', 'screenname', None)
    else:
        config.add_section('general')
        config.set('general', 'screenname', None)

    return (config_file, is_xdg, config)


def get_config(config=None, config_file=None) -> dict:
    _config = config or {}
    config_file, is_xdg, config = parse_config_file(config_file)

    if not os.path.isfile(config_file):
        return _config

    config_dict: dict = defaultdict(dict)
    for sec in config.sections():
        try:
            for k, v in config.items(sec):
                if k is None or v is None:
                    continue
                config_dict[sec][k] = v
        except TypeError:
            pass

    # Recursive/deep update.
    deep_update(config_dict, _config)

    return {k: v for k, v in config_dict.items() if v is not None}



"""
ia_move.py

'ia' subcommand for moving files on archive.org
"""

# Copyright (C) 2012-2024 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import sys

from internetarchive.cli import ia_copy
from internetarchive.cli.cli_utils import MetadataAction, QueryStringAction


def setup(subparsers):
    """
    Setup args for move command.

    Args:
        subparsers: subparser object passed from ia.py
    """
    parser = subparsers.add_parser("move",
                                   aliases=["mv"],
                                   help="Move and rename files in archive.org items")

    # Positional arguments
    parser.add_argument("source",
                        metavar="SOURCE",
                        help="Source file formatted as: identifier/file")
    parser.add_argument("destination",
                        metavar="DESTINATION",
                        help="Destination file formatted as: identifier/file")

    # Options
    parser.add_argument("-m", "--metadata",
                        metavar="KEY:VALUE",
                        nargs="+",
                        action=MetadataAction,
                        help=("Metadata to add to your new item, "
                              "if you are moving the file to a new item"))
    parser.add_argument("-H", "--header",
                        metavar="KEY:VALUE",
                        nargs="+",
                        action=QueryStringAction,
                        default={},
                        help="S3 HTTP headers to send with your request")
    parser.add_argument("--replace-metadata",
                        action="store_true",
                        help=("Only use metadata specified as argument, do not copy any "
                              "from the source item"))
    parser.add_argument("--ignore-file-metadata",
                        action="store_true",
                        help="Do not copy file metadata")
    parser.add_argument("-n", "--no-derive",
                        action="store_true",
                        help="Do not derive uploaded files")
    parser.add_argument("--no-backup",
                        action="store_true",
                        help=("Turn off archive.org backups, "
                              "clobbered files will not be saved to 'history/files/$key.~N~'"))

    parser.set_defaults(func=lambda args: main(args, parser))


def main(args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
    """
    Main entry point for ia move command.
    """
    # Add keep-old-version by default.
    if not args.header.get("x-archive-keep-old-version") and not args.no_backup:
        args.header["x-archive-keep-old-version"] = "1"

    # Call ia_copy.
    _, src_file = ia_copy.main(args, cmd="move", parser=parser)
    if src_file:
        dr = src_file.delete(headers=args.header, cascade_delete=True)
    else:
        print(f"error: {src_file} does not exist", file=sys.stderr)
        sys.exit(1)
    if dr.status_code == 204:
        print(f"success: moved '{args.source}' to '{args.destination}'", file=sys.stderr)
        sys.exit(0)
    print(f"error: {dr.content}", file=sys.stderr)



import sys
from time import time

import responses

from tests.conftest import IaRequestsMock, ia_call


def test_ia_metadata_exists(capsys):
    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        ia_call(['ia', 'metadata', '--exists', 'nasa'], expected_exit_code=0)
        out, err = capsys.readouterr()
        assert err == 'nasa exists\n'
        rsps.reset()
        rsps.add_metadata_mock('nasa', '{}')
        sys.argv = ['ia', 'metadata', '--exists', 'nasa']
        ia_call(['ia', 'metadata', '--exists', 'nasa'], expected_exit_code=1)
        out, err = capsys.readouterr()
        assert err == 'nasa does not exist\n'


def test_ia_metadata_formats(capsys, nasa_mocker):
    ia_call(['ia', 'metadata', '--formats', 'nasa'])
    out, err = capsys.readouterr()
    expected_formats = {'Collection Header', 'Archive BitTorrent', 'JPEG',
                        'Metadata', ''}
    assert set(out.split('\n')) == expected_formats


def test_ia_metadata_modify(capsys):
    md_rsp = ('{"success":true,"task_id":447613301,'
              '"log":"https://catalogd.archive.org/log/447613301"}')
    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa', method=responses.GET)
        rsps.add_metadata_mock('nasa', body=md_rsp, method=responses.POST)
        valid_key = f'foo-{int(time())}'
        ia_call(['ia', 'metadata', 'nasa', '--modify', f'{valid_key}:test_value'])
        out, err = capsys.readouterr()
        assert err == 'nasa - success: https://catalogd.archive.org/log/447613301\n'



import os
import sys
from contextlib import contextmanager
from io import StringIO

import responses

from internetarchive.utils import json
from tests.conftest import IaRequestsMock, ia_call, load_test_data_file

PROTOCOL = 'https:'
STATUS_CHECK_RESPONSE = load_test_data_file('s3_status_check.json')


def test_ia_upload(tmpdir_ch, caplog):
    with open('test.txt', 'w') as fh:
        fh.write('foo')

    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT, f'{PROTOCOL}//s3.us.archive.org/nasa/test.txt',
                 body='',
                 content_type='text/plain')
        ia_call(['ia', '--log', 'upload', 'nasa', 'test.txt'])

    assert f'uploaded test.txt to {PROTOCOL}//s3.us.archive.org/nasa/test.txt' in caplog.text


def test_ia_upload_invalid_identifier(capsys, caplog):
    with open('test.txt', 'w') as fh:
        fh.write('foo')

    ia_call(['ia', '--log', 'upload', 'føø', 'test.txt'],
            expected_exit_code=2)

    out, err = capsys.readouterr()
    assert "Identifier can only contain alphanumeric" in err


def test_ia_upload_status_check(capsys):
    with IaRequestsMock() as rsps:
        rsps.add(responses.GET, f'{PROTOCOL}//s3.us.archive.org',
                 body=STATUS_CHECK_RESPONSE,
                 content_type='application/json')

        ia_call(['ia', 'upload', 'nasa', '--status-check'])
        out, err = capsys.readouterr()
        assert 'success: nasa is accepting requests.' in err

        j = json.loads(STATUS_CHECK_RESPONSE)
        j['over_limit'] = 1
        rsps.reset()
        rsps.add(responses.GET, f'{PROTOCOL}//s3.us.archive.org',
                 body=json.dumps(j),
                 content_type='application/json')

        ia_call(['ia', 'upload', 'nasa', '--status-check'], expected_exit_code=1)
        out, err = capsys.readouterr()
        assert ('warning: nasa is over limit, and not accepting requests. '
                'Expect 503 SlowDown errors.') in err


def test_ia_upload_debug(capsys, tmpdir_ch, nasa_mocker):
    with open('test.txt', 'w') as fh:
        fh.write('foo')

    ia_call(['ia', 'upload', '--debug', 'nasa', 'test.txt'])
    out, err = capsys.readouterr()
    assert 'User-Agent' in err
    assert 's3.us.archive.org/nasa/test.txt' in err
    assert 'Accept:*/*' in err
    assert 'Authorization:LOW ' in err
    assert 'Connection:close' in err
    assert 'Content-Length:3' in err
    assert 'Accept-Encoding:gzip, deflate' in err


def test_ia_upload_403(capsys):
    s3_error = ('<Error>'
                '<Code>SignatureDoesNotMatch</Code>'
                '<Message>The request signature we calculated does not match '
                'the signature you provided. Check your AWS Secret Access Key '
                'and signing method. For more information, see REST '
                'Authentication and SOAP Authentication for details.</Message>'
                "<Resource>'PUT\n\n\n\n/iacli-test-item60/test-replace.txt'</Resource>"
                '<RequestId>18a9c5ea-088f-42f5-9fcf-70651cc085ca</RequestId>'
                '</Error>')

    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT,
                 f'{PROTOCOL}//s3.us.archive.org/nasa/test_ia_upload.py',
                 body=s3_error,
                 status=403,
                 content_type='text/plain')
        ia_call(['ia', 'upload', 'nasa', __file__], expected_exit_code=1)

    out, err = capsys.readouterr()
    assert 'error uploading test_ia_upload.py' in err


def test_ia_upload_invalid_cmd(capsys):
    ia_call(['ia', 'upload', 'nasa', 'nofile.txt'], expected_exit_code=2)
    out, err = capsys.readouterr()

    assert "'nofile.txt' is not a valid file or directory" in err


def test_ia_upload_size_hint(capsys, tmpdir_ch, nasa_mocker):
    with open('test.txt', 'w') as fh:
        fh.write('foo')

    ia_call(['ia', 'upload', '--debug', '--size-hint', '30', 'nasa', 'test.txt'])
    out, err = capsys.readouterr()
    assert 'User-Agent' in err
    assert 's3.us.archive.org/nasa/test.txt' in err
    assert 'x-archive-size-hint:30' in err
    assert 'Accept:*/*' in err
    assert 'Authorization:LOW ' in err
    assert 'Connection:close' in err
    assert 'Content-Length:3' in err
    assert 'Accept-Encoding:gzip, deflate' in err


def test_ia_upload_automatic_size_hint_files(capsys, tmpdir_ch, nasa_mocker):
    with open('foo', 'w') as fh:
        fh.write('foo')
    with open('bar', 'w') as fh:
        fh.write('bar')

    ia_call(['ia', 'upload', '--debug', 'nasa', 'foo', 'bar'])
    out, err = capsys.readouterr()
    assert 'x-archive-size-hint:6' in err

def test_ia_upload_automatic_size_hint_dir(capsys, tmpdir_ch, nasa_mocker):
    with open('foo', 'w') as fh:
        fh.write('foo')
    with open('bar', 'w') as fh:
        fh.write('bar')

    ia_call(['ia', 'upload', '--debug', 'nasa', '.'], expected_exit_code=2)
    out, err = capsys.readouterr()
    assert 'x-archive-size-hint:115' in err


def test_ia_upload_unicode(tmpdir_ch, caplog):
    with open('தமிழ் - baz ∆.txt', 'w') as fh:
        fh.write('unicode foo')

    efname = '%E0%AE%A4%E0%AE%AE%E0%AE%BF%E0%AE%B4%E0%AF%8D%20-%20baz%20%E2%88%86.txt'
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT,
                 f'{PROTOCOL}//s3.us.archive.org/nasa/{efname}',
                 body='',
                 content_type='text/plain')
        ia_call(['ia', '--log', 'upload', 'nasa', 'தமிழ் - baz ∆.txt',
                 '--metadata', 'foo:∆'])

    assert (f'uploaded தமிழ் - baz ∆.txt to {PROTOCOL}//s3.us.archive.org/nasa/'
            '%E0%AE%A4%E0%AE%AE%E0%AE%BF%E0%AE%B4%E0%AF%8D%20-%20'
            'baz%20%E2%88%86.txt') in caplog.text


def test_ia_upload_remote_name(tmpdir_ch, caplog):
    with open('test.txt', 'w') as fh:
        fh.write('foo')

    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT, f'{PROTOCOL}//s3.us.archive.org/nasa/hi.txt',
                 body='',
                 content_type='text/plain')
        ia_call(['ia', '--log', 'upload', 'nasa', 'test.txt', '--remote-name',
                 'hi.txt'])

    assert f'uploaded hi.txt to {PROTOCOL}//s3.us.archive.org/nasa/hi.txt' in caplog.text


def test_ia_upload_stdin(tmpdir_ch, caplog):
    @contextmanager
    def replace_stdin(f):
        original_stdin = sys.stdin
        sys.stdin = f
        try:
            yield
        finally:
            sys.stdin = original_stdin

    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT, f'{PROTOCOL}//s3.us.archive.org/nasa/hi.txt',
                 body='',
                 content_type='text/plain')
        with replace_stdin(StringIO('foo')):
            ia_call(['ia', '--log', 'upload', 'nasa', '-', '--remote-name', 'hi.txt'])

    assert f'uploaded hi.txt to {PROTOCOL}//s3.us.archive.org/nasa/hi.txt' in caplog.text


def test_ia_upload_inexistent_file(tmpdir_ch, capsys, caplog):
    ia_call(['ia', 'upload', 'foo', 'test.txt'], expected_exit_code=2)
    out, err = capsys.readouterr()
    assert "'test.txt' is not a valid file or directory" in err


def test_ia_upload_spreadsheet(tmpdir_ch, capsys):
    with open('foo.txt', 'w') as fh:
        fh.write('foo')
    with open('test.txt', 'w') as fh:
        fh.write('bar')
    with open('test.csv', 'w') as fh:
        fh.write('identifier,file,REMOTE_NAME\n')
        fh.write('nasa,foo.txt,\n')
        fh.write(',test.txt,bar.txt\n')

    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT, f'{PROTOCOL}//s3.us.archive.org/nasa/foo.txt',
                 body='',
                 content_type='text/plain')
        rsps.add(responses.PUT, f'{PROTOCOL}//s3.us.archive.org/nasa/bar.txt',
                 body='',
                 content_type='text/plain')
        ia_call(['ia', 'upload', '--spreadsheet', 'test.csv'])

    out, err = capsys.readouterr()
    assert 'uploading foo.txt' in err
    assert 'uploading bar.txt' in err


def test_ia_upload_spreadsheet_item_column(tmpdir_ch, capsys):
    with open('test.txt', 'w') as fh:
        fh.write('foo')
    with open('test.csv', 'w') as fh:
        fh.write('item,file\n')
        fh.write('nasa,test.txt\n')

    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT, f'{PROTOCOL}//s3.us.archive.org/nasa/test.txt',
                 body='',
                 content_type='text/plain')
        ia_call(['ia', 'upload', '--spreadsheet', 'test.csv'])

    out, err = capsys.readouterr()
    assert 'uploading test.txt' in err


def test_ia_upload_spreadsheet_item_and_identifier_column(tmpdir_ch, capsys):
    # item is preferred, and both are discarded
    with open('test.txt', 'w') as fh:
        fh.write('foo')
    with open('test.csv', 'w') as fh:
        fh.write('item,identifier,file\n')
        fh.write('nasa,uhoh,test.txt\n')

    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT, f'{PROTOCOL}//s3.us.archive.org/nasa/test.txt',
                 body='',
                 content_type='text/plain')

        ia_call(['ia', 'upload', '--spreadsheet', 'test.csv'])

        # Verify that the item and identifier columns are not in the PUT request headers
        putCalls = [c for c in rsps.calls if c.request.method == 'PUT']
        assert len(putCalls) == 1
        assert 'x-archive-meta00-identifier' not in putCalls[0].request.headers
        assert 'x-archive-meta00-item' not in putCalls[0].request.headers

    out, err = capsys.readouterr()
    assert 'uploading test.txt' in err


def test_ia_upload_spreadsheet_missing_identifier(tmpdir_ch, capsys, caplog):
    with open('test.txt', 'w') as fh:
        fh.write('foo')
    with open('test.csv', 'w') as fh:
        fh.write('file\n')
        fh.write('test.txt\n')

    ia_call(['ia', 'upload', '--spreadsheet', 'test.csv'], expected_exit_code=1)

    assert 'error: no identifier column on spreadsheet.' in capsys.readouterr().err


def test_ia_upload_spreadsheet_empty_identifier(tmpdir_ch, capsys, caplog):
    with open('test.txt', 'w') as fh:
        fh.write('foo')
    with open('test.csv', 'w') as fh:
        fh.write('identifier,file\n')
        fh.write(',test.txt\n')

    ia_call(['ia', 'upload', '--spreadsheet', 'test.csv'], expected_exit_code=1)

    assert 'error: no identifier column on spreadsheet.' in capsys.readouterr().err


def test_ia_upload_spreadsheet_bom(tmpdir_ch, capsys):
    with open('test.txt', 'w') as fh:
        fh.write('foo')
    with open('test.csv', 'wb') as fh:
        fh.write(b'\xef\xbb\xbf')
        fh.write(b'identifier,file\n')
        fh.write(b'nasa,test.txt\n')

    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT, f'{PROTOCOL}//s3.us.archive.org/nasa/test.txt',
                 body='',
                 content_type='text/plain')
        ia_call(['ia', 'upload', '--spreadsheet', 'test.csv'])

    out, err = capsys.readouterr()
    assert 'uploading test.txt' in err


def test_ia_upload_checksum(tmpdir_ch, caplog):
    with open('test.txt', 'w') as fh:
        fh.write('foo')

    # First upload, file not in metadata yet
    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT, f'{PROTOCOL}//s3.us.archive.org/nasa/test.txt',
                 body='',
                 content_type='text/plain')
        ia_call(['ia', '--log', 'upload', 'nasa', 'test.txt', '--checksum'])
    assert f'uploaded test.txt to {PROTOCOL}//s3.us.archive.org/nasa/test.txt' in caplog.text

    caplog.clear()

    # Second upload with file in metadata
    def insert_test_txt(body):
        body = json.loads(body)
        body['files'].append({'name': 'test.txt', 'md5': 'acbd18db4cc2f85cedef654fccc4a4d8'})
        return json.dumps(body)

    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa', transform_body=insert_test_txt)
        ia_call(['ia', '--log', 'upload', 'nasa', 'test.txt', '--checksum'], expected_exit_code=1)

    assert f'test.txt already exists: {PROTOCOL}//s3.us.archive.org/nasa/test.txt' in caplog.text

    caplog.clear()

    # Second upload with spreadsheet
    with open('test.csv', 'w') as fh:
        fh.write('identifier,file\n')
        fh.write('nasa,test.txt\n')

    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa', transform_body=insert_test_txt)
        ia_call(['ia', '--log', 'upload', '--spreadsheet', 'test.csv', '--checksum'],
                expected_exit_code=1)

    assert f'test.txt already exists: {PROTOCOL}//s3.us.archive.org/nasa/test.txt' in caplog.text


def test_ia_upload_keep_directories(tmpdir_ch, caplog):
    os.mkdir('foo')
    with open('foo/test.txt', 'w') as fh:
        fh.write('foo')
    with open('test.csv', 'w') as fh:
        fh.write('identifier,file\n')
        fh.write('nasa,foo/test.txt\n')

    # Default behaviour
    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT, f'{PROTOCOL}//s3.us.archive.org/nasa/test.txt',
                 body='',
                 content_type='text/plain')
        ia_call(['ia', '--log', 'upload', 'nasa', 'foo/test.txt'])
    assert f'uploaded test.txt to {PROTOCOL}//s3.us.archive.org/nasa/test.txt' in caplog.text
    caplog.clear()

    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT, f'{PROTOCOL}//s3.us.archive.org/nasa/test.txt',
                 body='',
                 content_type='text/plain')
        ia_call(['ia', '--log', 'upload', '--spreadsheet', 'test.csv'])
    assert f'uploaded test.txt to {PROTOCOL}//s3.us.archive.org/nasa/test.txt' in caplog.text
    caplog.clear()

    # With the option
    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT, f'{PROTOCOL}//s3.us.archive.org/nasa/foo/test.txt',
                 body='',
                 content_type='text/plain')
        ia_call(['ia', '--log', 'upload', 'nasa', 'foo/test.txt', '--keep-directories'])
    assert f'uploaded foo/test.txt to {PROTOCOL}//s3.us.archive.org/nasa/foo/test.txt' in caplog.text
    caplog.clear()

    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('nasa')
        rsps.add(responses.PUT, f'{PROTOCOL}//s3.us.archive.org/nasa/foo/test.txt',
                 body='',
                 content_type='text/plain')
        ia_call(['ia', '--log', 'upload', '--spreadsheet', 'test.csv', '--keep-directories'])
    assert f'uploaded foo/test.txt to {PROTOCOL}//s3.us.archive.org/nasa/foo/test.txt' in caplog.text



from tests.conftest import ia_call


def test_ia(capsys):
    ia_call(['ia', '--help'])
    out, err = capsys.readouterr()
    assert 'A command line interface to Archive.org.' in out

    ia_call(['ia', '--insecure', 'ls', 'nasa'])

    ia_call(['ia', 'nocmd'], expected_exit_code=2)
    out, err = capsys.readouterr()
    assert "invalid choice: 'nocmd'" in err



"""
ia_account.py

'ia' subcommand for configuring 'ia' with your archive.org credentials.
"""

# Copyright (C) 2012-2025 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import sys

from internetarchive import configure
from internetarchive.account import Account
from internetarchive.exceptions import AccountAPIError
from internetarchive.utils import is_valid_email


def setup(subparsers):
    """
    Setup args for configure command.

    Args:
        subparsers: subparser object passed from ia.py
    """
    parser = subparsers.add_parser("account",
                                   aliases=["ac"],
                                   description=(
                                       "Manage an archive.org account.\n\n"
                                       "Note: This command requires administrative "
                                       "privileges. "
                                   ),
                                   help=("Manage an archive.org account. "
                                         "Note: requires admin privileges"))

    group = parser.add_mutually_exclusive_group()
    parser.add_argument("user",
                        help="Email address, screenname, or itemname "
                             "for an archive.org account")
    group.add_argument("-g", "--get-email",
                        action="store_true",
                        help="Print the email address associated with the user and exit")
    group.add_argument("-s", "--get-screenname",
                        action="store_true",
                        help="Print the screenname associated with the user and exit")
    group.add_argument("-i", "--get-itemname",
                        action="store_true",
                        help="Print the itemname associated with the user and exit")
    group.add_argument("-l", "--is-locked",
                        action="store_true",
                        help="Check if an account is locked")
    group.add_argument("-L", "--lock",
                        action="store_true",
                        help="Lock an account")
    group.add_argument("-u", "--unlock",
                        action="store_true",
                        help="Unlock an account")

    parser.add_argument("-c", "--comment",
                        type=str,
                        help="Comment to include with lock/unlock action")

    parser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    """
    Main entrypoint for 'ia account'.
    """
    try:
        if args.user.startswith('@'):
            account = Account.from_account_lookup('itemname', args.user)
        elif not is_valid_email(args.user):
            account = Account.from_account_lookup('screenname', args.user)
        else:
            account = Account.from_account_lookup('email', args.user)
    except AccountAPIError as exc:
        print(json.dumps(exc.error_data))
        sys.exit(1)

    if args.get_email:
        print(account.canonical_email)
    elif args.get_screenname:
        print(account.screenname)
    elif args.get_itemname:
        print(account.itemname)
    elif args.is_locked:
        print(account.locked)
    elif args.lock:
        r = account.lock(args.comment, session=args.session)
        print(r.text)
    elif args.unlock:
        r = account.unlock(args.comment, session=args.session)
        print(r.text)
    else:
        account_data = account.to_dict()
        print(json.dumps(account_data))



"""
ia_reviews.py

'ia' subcommand for listing, submitting, and deleting reviews for archive.org items.
"""

# Copyright (C) 2012-2024 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import sys

from requests.exceptions import HTTPError


def setup(subparsers):
    """
    Setup args for list command.

    Args:
        subparsers: subparser object passed from ia.py
    """
    parser = subparsers.add_parser("reviews",
                                   aliases=["re"],
                                   help="submit and modify reviews for archive.org items")

    # Positional arguments
    parser.add_argument("identifier",
                        type=str,
                        help="identifier of the item")

    # Options
    parser.add_argument("-d", "--delete",
                        action="store_true",
                        help="delete your review")
    parser.add_argument("-t", "--title",
                        type=str,
                        help="the title of your review")
    parser.add_argument("-b", "--body",
                        type=str,
                        help="the body of your review")
    parser.add_argument("-s", "--stars",
                        type=int,
                        help="the number of stars for your review")
    parser.add_argument("-i", "--index",
                        action="store_true",
                        help="Index a review")
    parser.add_argument("-n", "--noindex",
                        action="store_true",
                        help="Remove a review from the index")

    # Conditional arguments that require --delete
    delete_group = parser.add_argument_group("delete options",
                                             ("these options are used with "
                                              "the --delete flag"))
    delete_group.add_argument("-u", "--username",
                              type=str,
                              help="delete reviews for a specific user given USERNAME")
    delete_group.add_argument("-S", "--screenname",
                              type=str,
                              help="delete reviews for a specific user given SCREENNAME")
    delete_group.add_argument("-I", "--itemname",
                              type=str,
                              help="delete reviews for a specific user given ITEMNAME")

    parser.set_defaults(func=lambda args: main(args, parser))


def main(args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
    """
    Main entry point for 'ia reviews'.
    """
    item = args.session.get_item(args.identifier)
    if args.index:
        r = item.index_review(username=args.username,
                               screenname=args.screenname,
                               itemname=args.itemname)
        if r.json().get("success"):
            print(f"{item.identifier} - success: review indexed", file=sys.stderr)
            sys.exit(0)
    elif args.noindex:
        r = item.noindex_review(username=args.username,
                                screenname=args.screenname,
                                itemname=args.itemname)
        if r.json().get("success"):
            print(f"{item.identifier} - success: review removed from index", file=sys.stderr)
            sys.exit(0)
    if args.delete:
        r = item.delete_review(username=args.username,
                               screenname=args.screenname,
                               itemname=args.itemname)
    elif not args.body and not args.title:
        try:
            r = item.get_review()
            print(r.text)
            sys.exit(0)
        except HTTPError as exc:
            if exc.response.status_code == 404:  # type: ignore
                sys.exit(0)
            else:
                raise exc
    else:
        if (args.title and not args.body) or (args.body and not args.title):
            parser.error("both --title and --body must be provided")
        r = item.review(args.title, args.body, args.stars)
    j = r.json()
    if j.get("success") or "no change detected" in j.get("error", "").lower():
        task_id = j.get("value", {}).get("task_id")
        if task_id:
            print((f"{item.identifier} - success: "
                   f"https://catalogd.archive.org/log/{task_id}"),
                  file=sys.stderr)
        else:
            print(f"{item.identifier} - warning: no changes detected!", file=sys.stderr)
        sys.exit(0)
    else:
        print(f"{item.identifier} - error: {j.get('error')}", file=sys.stderr)
        sys.exit(1)



import json
from dataclasses import dataclass, field
from typing import ClassVar, Dict, List, Optional

import requests

from internetarchive import get_session
from internetarchive.exceptions import AccountAPIError
from internetarchive.session import ArchiveSession

"""
internetarchive.account
~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (C) 2012-2025 by Internet Archive.
:license: AGPL 3, see LICENSE for more details.

This module provides the `Account` class for interacting with user accounts on the
Internet Archive. It requires administrative privileges.
"""


@dataclass
class Account:
    """
    A class for interacting with user accounts on the Internet Archive.

    Note:
        This class requires administrative privileges.

    This class provides methods to:
    - Fetch account details using various identifiers (e.g., email, screenname, itemname).
    - Lock and unlock accounts.
    - Convert account data to a dictionary for serialization.

    Example Usage:
        >>> from internetarchive.account import Account
        >>> account = Account.from_account_lookup('email', 'foo@example.com')
        >>> account.lock(comment="Locked spam account")
        >>> print(account.to_dict())
    """
    locked: bool
    verified: bool
    email: str
    canonical_email: str
    itemname: str
    screenname: str
    notifications: List[str]
    has_disability_access: bool
    lastlogin: str
    createdate: str
    session: ArchiveSession = field(default_factory=get_session)

    API_BASE_URL: str = '/services/xauthn/'
    API_INFO_PARAMS: ClassVar[Dict[str, str]] = {'op': 'info'}
    API_LOCK_UNLOCK_PARAMS: ClassVar[Dict[str, str]] = {'op': 'lock_unlock'}

    def _get_api_base_url(self) -> str:
        """Dynamically construct the API base URL using the session's host."""
        return f'https://{self.session.host}{self.API_BASE_URL}'  # type: ignore[attr-defined]

    def _post_api_request(
        self,
        endpoint: str,
        params: Dict[str, str],
        data: Dict[str, str],
        session: Optional[ArchiveSession] = None
    ) -> requests.Response:
        """
        Helper method to make API requests.

        Args:
            endpoint: The API endpoint to call.
            params: Query parameters for the request.
            data: Data to send in the request body.
            session: Optional session to use for the request. Defaults to self.session.

        Returns:
            The response from the API.

        Raises:
            requests.exceptions.RequestException: If the API request fails.
        """
        session = session or self.session
        url = f'https://{session.host}{endpoint}'  # type: ignore[attr-defined]
        response = session.post(url, params=params, data=data)
        response.raise_for_status()
        return response

    @classmethod
    def from_account_lookup(
        cls,
        identifier_type: str,
        identifier: str,
        session: Optional[ArchiveSession] = None
    ) -> "Account":
        """
        Factory method to initialize an Account using an identifier type and value.

        Args:
            identifier_type: The type of identifier (e.g., 'email', 'screenname').
            identifier: The value of the identifier (e.g., 'foo@example.com').
            session: Optional session to use for the request.

        Returns:
            An instance of Account.
        """
        json_data = cls._fetch_account_data_from_api(identifier_type, identifier, session)
        return cls.from_json(json_data, session)

    @classmethod
    def _fetch_account_data_from_api(
        cls,
        identifier_type: str,
        identifier: str,
        session: Optional[ArchiveSession] = None
    ) -> Dict:
        """
        Fetches account data from the API using an identifier type and value.

        Args:
            identifier_type: The type of identifier (e.g., 'email', 'screenname').
            identifier: The value of the identifier (e.g., 'foo@example.com').
            session: Optional session to use for the request.

        Returns:
            A dictionary containing the account data.

        Raises:
            requests.exceptions.RequestException: If the API request fails.
            ValueError: If the API response is invalid or missing required data.
        """
        data = {identifier_type: identifier}
        session = session or get_session()
        try:
            response = session.post(
                f'https://{session.host}{cls.API_BASE_URL}',  # type: ignore[attr-defined]
                params=cls.API_INFO_PARAMS,
                data=json.dumps(data)
            )
            response.raise_for_status()
            j = response.json()
            if j.get("error") or not j.get("values"):
                raise AccountAPIError(j.get("error", "Unknown error"), error_data=j)
            return j["values"]
        except requests.exceptions.RequestException as e:
            raise AccountAPIError(f"Failed to fetch account data: {e}")


    @classmethod
    def from_json(
        cls,
        json_data: Dict,
        session: Optional[ArchiveSession] = None
    ) -> "Account":
        """
        Factory method to initialize an Account using JSON data.

        Args:
            json_data: A dictionary containing account data.
            session: Optional session to use for the request.

        Returns:
            An instance of Account.

        Raises:
            ValueError: If required fields are missing in the JSON data.
        """
        required_fields = [
            "canonical_email",
            "email",
            "has_disability_access",
            "itemname",
            "locked",
            "notifications",
            "screenname",
            "verified",
            "lastlogin",
            "createdate",
        ]
        for requried_field in required_fields:
            if requried_field not in json_data:
                raise ValueError(f"Missing required requried_field in JSON data: {requried_field}")

        # Ensure session is of type ArchiveSession
        if session is None:
            session = get_session()  # Default to ArchiveSession
        elif not isinstance(session, ArchiveSession):
            raise TypeError(f"Expected session to be of type ArchiveSession, got {type(session)}")

        return cls(
            locked=json_data["locked"],
            verified=json_data["verified"],
            email=json_data["email"],
            canonical_email=json_data["canonical_email"],
            itemname=json_data["itemname"],
            screenname=json_data["screenname"],
            notifications=json_data["notifications"],
            has_disability_access=json_data["has_disability_access"],
            lastlogin=json_data["lastlogin"],
            createdate=json_data["createdate"],
            session=session
        )

    def lock(self,
             comment: Optional[str] = None,
             session: Optional[ArchiveSession] = None) -> requests.Response:
        """
        Lock the account.

        Args:
            comment: An optional comment for the lock operation.
            session: Optional session to use for the request.

        Returns:
            The response from the API.
        """
        data = {'itemname': self.itemname, 'is_lock': '1'}
        if comment:
            data['comments'] = comment
        return self._post_api_request(
            self.API_BASE_URL,
            params=self.API_LOCK_UNLOCK_PARAMS,
            data=data,
            session=session
        )

    def unlock(self,
               comment: Optional[str] = None,
               session: Optional[ArchiveSession] = None) -> requests.Response:
        """
        Unlock the account.

        Args:
            comment: An optional comment for the unlock operation.
            session: Optional session to use for the request.

        Returns:
            The response from the API.
        """
        data = {'itemname': self.itemname, 'is_lock': '0'}
        if comment:
            data['comments'] = comment
        return self._post_api_request(
            self.API_BASE_URL,
            params=self.API_LOCK_UNLOCK_PARAMS,
            data=data,
            session=session
        )

    def to_dict(self) -> Dict:
        """
        Converts the Account instance to a dictionary.

        Returns:
            A dictionary representation of the Account instance.
        """
        return {
            "locked": self.locked,
            "verified": self.verified,
            "email": self.email,
            "canonical_email": self.canonical_email,
            "itemname": self.itemname,
            "screenname": self.screenname,
            "notifications": self.notifications,
            "has_disability_access": self.has_disability_access,
            "lastlogin": self.lastlogin,
            "createdate": self.createdate,
        }
